#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from warnings import catch_warnings

import numpy as np
from ax.exceptions.core import UnsupportedError
from ax.exceptions.model import ModelError
from ax.generators.discrete.thompson import ThompsonSampler
from ax.utils.common.testutils import TestCase


class ThompsonSamplerTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Xs = [[[1, 1], [2, 2], [3, 3], [4, 4]]]  # 4 arms, each of dimensionality 2
        self.Ys = [[1, 2, 3, 4]]
        self.Yvars = [[1, 1, 1, 1]]
        self.parameter_values = [[1, 2, 3, 4], [1, 2, 3, 4]]
        self.outcome_names = ["x", "y"]  # not used for regular TS

        self.multiple_metrics_Xs = [
            [[1, 1], [2, 2], [3, 3], [4, 4]],
            [[1, 1], [2, 2], [3, 3], [4, 4]],
        ]  # 2 metrics, 4 arms, each of dimensionality 2
        self.multiple_metrics_Ys = [[1, 2, 3, 4], [0, 0, 0, 1]]
        self.multiple_metrics_Yvars = [[1, 1, 1, 1], [1, 1, 1, 1]]

    def test_ThompsonSampler(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        self.assertFalse(generator.can_predict)
        self.assertFalse(generator.can_model_in_sample)
        self.assertEqual(generator.topk, 1)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        arms, weights, gen_metadata = generator.gen(
            n=3,
            parameter_values=self.parameter_values,
            objective_weights=np.ones(1),
        )
        self.assertEqual(arms, [[4, 4], [3, 3], [2, 2]])
        for weight, expected_weight in zip(
            weights, [3 * i for i in [0.725, 0.225, 0.05]]
        ):
            self.assertAlmostEqual(weight, expected_weight, 1)
        self.assertEqual(len(gen_metadata["arms_to_weights"]), 4)
        self.assertEqual(gen_metadata["best_x"], arms[0])

    def test_ThompsonSamplerValidation(self) -> None:
        generator = ThompsonSampler(min_weight=0.01)

        # all Xs are not the same
        with self.assertRaises(ValueError):
            generator.fit(
                Xs=[[[1, 1], [2, 2], [3, 3], [4, 4]], [[1, 1], [2, 2], [4, 4]]],
                Ys=self.Ys,
                Yvars=self.Yvars,
                parameter_values=self.parameter_values,
                outcome_names=self.outcome_names,
            )

        # multiple observations per parameterization
        with self.assertRaises(ValueError):
            generator.fit(
                Xs=[[[1, 1], [2, 2], [2, 2]]],
                Ys=self.Ys,
                Yvars=self.Yvars,
                parameter_values=self.parameter_values,
                outcome_names=self.outcome_names,
            )

        # these are not the same observations, so should not error
        generator.fit(
            Xs=[[[1, 1], [2.0, 2], [2, 2]]],
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )

        # requires objective weights
        with self.assertRaises(ValueError):
            generator.gen(5, self.parameter_values, objective_weights=None)

    def test_ThompsonSamplerTopKError(self) -> None:
        generator = ThompsonSampler(topk=5)
        with self.assertRaisesRegex(ModelError, r"ThompsonSampler `topk=\d+`"):
            generator.fit(
                Xs=self.Xs,
                Ys=self.Ys,
                Yvars=self.Yvars,
                parameter_values=self.parameter_values,
                outcome_names=self.outcome_names,
            )

    def test_TopTwo_alters_weights_vs_TopOne(self) -> None:
        np.random.seed(0)

        # Compare TTTS results to the vanilla TS
        ts1 = ThompsonSampler(num_samples=1000, min_weight=0.0, topk=1)
        ts1.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        _, _, gen_metadata1 = ts1.gen(
            n=4,
            parameter_values=self.parameter_values,
            objective_weights=np.ones(1),
        )
        full_w1 = gen_metadata1["arms_to_weights"]

        ts2 = ThompsonSampler(num_samples=1000, min_weight=0.0, topk=2)
        ts2.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        _, _, gen_metadata2 = ts2.gen(
            n=4,
            parameter_values=self.parameter_values,
            objective_weights=np.ones(1),
        )
        full_w2 = gen_metadata2["arms_to_weights"]

        # Sanity: same length
        self.assertEqual(len(full_w1), 4)
        self.assertEqual(len(full_w2), 4)

        # 1) Best arm (4) weight must drop under TTTS
        self.assertLess(full_w2[3], full_w1[3])

        # 2) Second-best arm (3) weight must rise
        self.assertGreater(full_w2[2], full_w1[2])

        # 3) Arm 2 (index 1) and arm 1 (index 0) both see P(2nd)>P(best)
        #   so they increase
        self.assertGreater(full_w2[1], full_w1[1])
        self.assertGreater(full_w2[0], full_w1[0])

        # 4) Monotonicity in the final TTTS distribution still holds
        self.assertTrue(full_w2[3] > full_w2[2] > full_w2[1] > full_w2[0])

    def test_ThompsonSamplerMinWeight(self) -> None:
        np.random.seed(0)
        generator = ThompsonSampler(min_weight=0.01)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        arms, weights, _ = generator.gen(
            n=3,
            parameter_values=self.parameter_values,
            objective_weights=np.ones(1),
        )
        self.assertEqual(arms, [[4, 4], [3, 3], [2, 2]])
        for weight, expected_weight in zip(
            weights, [3 * i for i in [0.725, 0.225, 0.05]]
        ):
            self.assertAlmostEqual(weight, expected_weight, 1)

    def test_ThompsonSamplerUniformWeights(self) -> None:
        generator = ThompsonSampler(min_weight=0.0, uniform_weights=True)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        arms, weights, _ = generator.gen(
            n=3,
            parameter_values=self.parameter_values,
            objective_weights=np.ones(1),
        )
        self.assertEqual(arms, [[4, 4], [3, 3], [2, 2]])
        for weight, expected_weight in zip(weights, [1.0, 1.0, 1.0]):
            self.assertAlmostEqual(weight, expected_weight, 1)

    def test_ThompsonSamplerInfeasible(self) -> None:
        generator = ThompsonSampler(min_weight=0.9)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        with self.assertRaises(ModelError):
            generator.gen(
                n=3,
                parameter_values=self.parameter_values,
                objective_weights=np.ones(1),
            )

    def test_ThompsonSamplerOutcomeConstraints(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        generator.fit(
            Xs=self.multiple_metrics_Xs,
            Ys=self.multiple_metrics_Ys,
            Yvars=self.multiple_metrics_Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        arms, weights, _ = generator.gen(
            n=4,
            parameter_values=self.parameter_values,
            objective_weights=np.array([1, 0]),
            outcome_constraints=(
                # pass in multiples of the same constraint
                # to ensure that shapes are correct for multiple constraints
                np.array([[0, 1], [0, 1], [0, 1]]),
                np.array([[1], [1], [1]]),
            ),
        )
        self.assertEqual(arms, [[3, 3], [4, 4], [2, 2], [1, 1]])
        for weight, expected_weight in zip(
            weights, [4 * i for i in [0.4, 0.4, 0.15, 0.05]]
        ):
            self.assertAlmostEqual(weight, expected_weight, delta=0.15)

    def test_ThompsonSamplerOutcomeConstraintsInfeasible(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        generator.fit(
            Xs=self.multiple_metrics_Xs,
            Ys=self.multiple_metrics_Ys,
            Yvars=self.multiple_metrics_Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        with self.assertRaises(ModelError):
            generator.gen(
                n=3,
                parameter_values=self.parameter_values,
                objective_weights=np.ones(2),
                outcome_constraints=(np.array([[0, 1]]), np.array([[-10]])),
            )

    def test_ThompsonSamplerPredict(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        f, cov = generator.predict([[1, 1], [3, 3]])
        self.assertTrue(np.array_equal(f, np.array([[1], [3]])))
        self.assertTrue(np.array_equal(cov, np.ones((2, 1, 1))))

        with self.assertRaisesRegex(UnsupportedError, "out-of-sample"):
            generator.predict([[1, 2]])

    def test_ThompsonSamplerMultiObjectiveWarning(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        generator.fit(
            Xs=self.multiple_metrics_Xs,
            Ys=self.multiple_metrics_Ys,
            Yvars=self.multiple_metrics_Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        with catch_warnings(record=True) as warning_list:
            arms, weights, _ = generator.gen(
                n=4,
                parameter_values=self.parameter_values,
                objective_weights=np.array([1, -1]),
                outcome_constraints=None,
            )
        self.assertEqual(
            "In case of multi-objective adding metric values together might"
            " not lead to a meaningful result.",
            str(warning_list[0].message),
        )

    def test_ThompsonSamplerNonPositiveN(self) -> None:
        generator = ThompsonSampler(min_weight=0.0)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        for n in (-1, 0):
            with self.assertRaisesRegex(ValueError, "ThompsonSampler requires n > 0"):
                generator.gen(
                    n=n,
                    parameter_values=self.parameter_values,
                    objective_weights=np.ones(1),
                )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from ax.adapter.base import Adapter
from ax.analysis.analysis import Analysis

from ax.analysis.healthcheck.healthcheck_analysis import (
    create_healthcheck_analysis_card,
    HealthcheckAnalysisCard,
    HealthcheckStatus,
)
from ax.core.auxiliary import AuxiliaryExperimentPurpose
from ax.core.experiment import Experiment
from ax.exceptions.core import AxError, UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.stats.no_effects import check_experiment_effects_per_metric
from pyre_extensions import override


class TestOfNoEffectAnalysis(Analysis):
    """
    Analysis for checking whether a randomization test can show that there are any
    effects whatsoever. This test is performed independently on each metric
    and it is based on the Welch's test for testing whether the means across
    groups are identical assuming unequal variances across groups.
    """

    def __init__(self, no_effect_alpha: float = 0.05) -> None:
        r"""
        Args:
            no_effect_alpha: The confidence level at which to reject the
                null hypothesis of no experimental effects.

        """
        self.no_effect_alpha = no_effect_alpha

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> HealthcheckAnalysisCard:
        r"""
        Compute the test of no effect separately for all metrics in the
        experiment. If objective metrics are found to not have effects,
        we return a card with a warning status and message.

        Args:
            experiment: Ax experiment.
            generation_strategy: Ax generation strategy.
            adapter: Ax adapter adapter

        Returns:
            A HealthcheckAnalysisCard object deatailing which metrics we don't
            detect any effect for.

        """
        status = HealthcheckStatus.PASS
        subtitle = "Effects are observed for all objective metrics."
        title_status = "Success"

        if experiment is None:
            raise UserInputError("TestOfNoEffectAnalysis requires an Experiment.")
        if (
            pe_experiments := experiment.auxiliary_experiments_by_purpose.get(
                AuxiliaryExperimentPurpose.PE_EXPERIMENT
            )
        ) is not None and len(pe_experiments) > 0:
            objective_names = pe_experiments[0].experiment.parameters.keys()
        else:
            if (opt_config := experiment.optimization_config) is not None:
                objective_names = opt_config.objective.metric_names
            else:
                raise UserInputError(
                    (
                        "TestOfNoEffectAnalysis requires an optimization config "
                        "to be set on the experiment."
                    )
                )

        data = experiment.lookup_data()
        if data.df.empty:
            raise AxError(
                (
                    "TestOfNoEffectAnalysis requires data to be attached "
                    "to the experiment. "
                )
            )

        df_tone = check_experiment_effects_per_metric(
            data=data, objective_names=set(objective_names)
        )
        metrics_tone = df_tone.groupby("metric_name")["has_effect"].sum() > 0
        metrics_with_effects = [i for i in metrics_tone.index if metrics_tone[i]]
        objectives_without_effects = set()
        if "is_objective" in df_tone.columns:
            objectives_with_data = set(df_tone[df_tone["is_objective"]]["metric_name"])
            objectives_without_effects = objectives_with_data.difference(
                set(metrics_with_effects)
            )

        if len(objectives_without_effects) > 0:
            status = HealthcheckStatus.WARNING
            formatted_metrics = "<br>".join(objectives_without_effects)
            subtitle = (
                "The test of no effect checks to see whether the objective "
                "metrics of an experiment have had any detectable effects. Based "
                "on the experiment data thus far, no effects have been detected at a "
                f"{(1.0 - self.no_effect_alpha) * 100:.0f}% confidence level for "
                f"the following metrics: <br><br> {formatted_metrics}"
            )
            title_status = "Warning"

        return create_healthcheck_analysis_card(
            name=self.__class__.__name__,
            title=f"Ax Test of No Effect {title_status}",
            subtitle=subtitle,
            df=df_tone,
            status=status,
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections import defaultdict
from collections.abc import Callable, Iterable
from logging import Logger
from typing import NamedTuple
from warnings import warn

import numpy as np
import numpy.typing as npt
from ax.adapter.base import Adapter, unwrap_observation_data
from ax.adapter.data_utils import ExperimentData
from ax.core.observation import Observation, ObservationData, ObservationFeatures
from ax.core.optimization_config import OptimizationConfig
from ax.exceptions.core import UnsupportedError
from ax.exceptions.model import CrossValidationError
from ax.utils.common.logger import get_logger
from ax.utils.stats.model_fit_stats import (
    coefficient_of_determination,
    compute_model_fit_metrics,
    DIAGNOSTIC_FNS,
    FISHER_EXACT_TEST_P,
    mean_of_the_standardized_error,
    ModelFitMetricProtocol,
    std_of_the_standardized_error,
)
from botorch.settings import validate_input_scaling
from pyre_extensions import assert_is_instance

logger: Logger = get_logger(__name__)

CVDiagnostics = dict[str, dict[str, float]]


class CVResult(NamedTuple):
    """Container for cross validation results."""

    observed: Observation
    predicted: ObservationData


class CVData(NamedTuple):
    """Data for cross validation."""

    training_data: ExperimentData
    test_data: ExperimentData


class AssessModelFitResult(NamedTuple):
    """Container for model fit assessment results"""

    good_fit_metrics_to_fisher_score: dict[str, float]
    bad_fit_metrics_to_fisher_score: dict[str, float]


def cross_validate(
    model: Adapter,
    folds: int = -1,
    test_selector: Callable[[Observation], bool] | None = None,
    untransform: bool = True,
    use_posterior_predictive: bool = False,
    fold_generator: Callable[[ExperimentData], Iterable[CVData]] | None = None,
) -> list[CVResult]:
    """Cross validation for model predictions.

    Splits the model's training data into train/test folds and makes
    out-of-sample predictions on the test folds.

    By default, train/test splits are made based on arm names, so that
    repeated observations of a arm will always be in the train or test set
    together. Different behavior can be achieved by passing in a custom
    fold_generator.

    The test set can be limited to a specific set of observations by passing in
    a test_selector callable. This function should take in an Observation
    and return a boolean indicating if it should be used in the test set or
    not. For example, we can limit the test set to arms with trial 0 with
    test_selector = lambda obs: obs.features.trial_index == 0
    If not provided, all observations will be available for the test set.

    Args:
        model: Fitted model (Adapter) to cross validate.
        folds: Number of folds. Use -1 for leave-one-out, otherwise will be
            k-fold. Unless fold_generator is used to specify different behavior.
        test_selector: Function for selecting observations for the test set.
        untransform: Whether to untransform the model predictions before
            cross validating.
            Generators are trained on transformed data, and candidate generation
            is performed in the transformed space. Computing the model
            quality metric based on the cross-validation results in the
            untransformed space may not be representative of the model that
            is actually used for candidate generation in case of non-invertible
            transforms, e.g., Winsorize or LogY.
            While the model in the transformed space may not be representative
            of the original data in regions where outliers have been removed,
            we have found it to better reflect the how good the model used
            for candidate generation actually is.
        use_posterior_predictive: A boolean indicating if the predictions
            should be from the posterior predictive (i.e. including
            observation noise). Note: we should reconsider how we compute
            cross-validation and model fit metrics where there is non-
            Gaussian noise.
        fold_generator: A function that generates train/test folds in the form
            of CVData objects. Defaults to k-fold CV.

    Returns:
        A CVResult for each observation in the training data.
    """
    # Get in-design training data.
    training_data = model.get_training_data(filter_in_design=True)
    if fold_generator is None:

        def fold_generator(training_data: ExperimentData) -> Iterable[CVData]:
            return _kfold_train_test_split(folds=folds, training_data=training_data)

    result = []
    for cv_data in fold_generator(training_data):
        cv_training_data = cv_data.training_data
        cv_test_data = cv_data.test_data
        cv_test_observations = [
            obs
            for obs in cv_test_data.convert_to_list_of_observations()
            if test_selector is None or test_selector(obs)
        ]
        cv_test_points = [obs.features for obs in cv_test_observations]
        if len(cv_test_points) == 0:
            continue

        # Make the prediction
        if untransform:
            cv_test_predictions = model.cross_validate(
                cv_training_data=cv_training_data,
                cv_test_points=cv_test_points,
                use_posterior_predictive=use_posterior_predictive,
            )
        else:
            # Get test predictions in transformed space
            (
                cv_training_data,
                cv_test_points,
                search_space,
            ) = model._transform_inputs_for_cv(
                cv_training_data=cv_training_data, cv_test_points=cv_test_points
            )
            # Since each CV fold removes points from the training data, the
            # remaining observations will not pass the input scaling checks.
            # To avoid confusing users with warnings, we disable these checks.
            with validate_input_scaling(False):
                cv_test_predictions = model._cross_validate(
                    search_space=search_space,
                    cv_training_data=cv_training_data,
                    cv_test_points=cv_test_points,
                    use_posterior_predictive=use_posterior_predictive,
                )
            # Get test observations in transformed space.
            for t in model.transforms.values():
                cv_test_data = t.transform_experiment_data(experiment_data=cv_test_data)
            # Re-construct the test observations with the transformed data.
            cv_test_observations = [
                obs
                for obs in cv_test_data.convert_to_list_of_observations()
                if test_selector is None or test_selector(obs)
            ]
        # Form CVResult objects
        if len(cv_test_observations) < len(cv_test_predictions):
            msg = (
                "There are fewer test observations than predictions. "
                "This can happen when transforms that reduce the number of "
                "observations are used in the Adapter used in cross validation. "
            )
            if folds == -1:
                msg += (
                    "Since this is leave-one-out cross validation, all observations "
                    "correspond to the same arm and we can utilize the first "
                    "observation in CV results."
                )
                logger.warning(msg)
            else:
                msg += (
                    "Since this is not leave-one-out cross validation, we cannot "
                    "guarantee consistency of predictions and observations. "
                    "Use leave-one-out cross validation with data reducing transforms, "
                    "or use cross validation with `untransform=True`."
                )
                raise CrossValidationError(msg)

        for observed, prediction in zip(
            cv_test_observations, cv_test_predictions, strict=False
        ):
            result.append(CVResult(observed=observed, predicted=prediction))
    return result


def compute_diagnostics(result: list[CVResult]) -> CVDiagnostics:
    """Computes diagnostics for given cross validation results.

    It provides a dictionary with values for the following diagnostics, for
    each metric:

    - 'Mean prediction CI': the average width of the CIs at each of the CV
      predictions, relative to the observed mean.
    - 'MAPE': mean absolute percentage error of the estimated mean relative
      to the observed mean.
    - 'wMAPE': Weighted mean absolute percentage error.
    - 'Total raw effect': the multiple change from the smallest observed
      mean to the largest observed mean, i.e. `(max - min) / min`.
    - 'Correlation coefficient': the Pearson correlation of the estimated
      and observed means.
    - 'Rank correlation': the Spearman correlation of the estimated
      and observed means.
    - 'Fisher exact test p': we test if the model is able to distinguish the
      bottom half of the observations from the top half, using Fisher's
      exact test and the observed/estimated means. A low p value indicates
      that the model has some ability to identify good arms. A high p value
      indicates that the model cannot identify arms better than chance, or
      that the observations are too noisy to be able to tell.

    Each of these is returned as a dictionary from metric name to value for
    that metric.

    Args:
        result: Output of cross_validate

    Returns:
        A dictionary keyed by diagnostic name with results as described above.
    """
    # Extract per-metric outcomes from CVResults.
    y_obs = defaultdict(list)
    y_pred = defaultdict(list)
    se_pred = defaultdict(list)
    for res in result:
        for j, metric_name in enumerate(res.observed.data.metric_names):
            y_obs[metric_name].append(res.observed.data.means[j])
            # Find the matching prediction
            k = res.predicted.metric_names.index(metric_name)
            y_pred[metric_name].append(res.predicted.means[k])
            se_pred[metric_name].append(np.sqrt(res.predicted.covariance[k, k]))
    y_obs = _arrayify_dict_values(y_obs)
    y_pred = _arrayify_dict_values(y_pred)
    se_pred = _arrayify_dict_values(se_pred)

    diagnostics = compute_model_fit_metrics(
        y_obs=y_obs,
        y_pred=y_pred,
        se_pred=se_pred,
        fit_metrics_dict=DIAGNOSTIC_FNS,
    )
    return diagnostics


def _arrayify_dict_values(d: dict[str, list[float]]) -> dict[str, npt.NDArray]:
    """Helper to convert dictionary values to numpy arrays."""
    return {k: np.array(v) for k, v in d.items()}


def assess_model_fit(
    diagnostics: CVDiagnostics,
    significance_level: float = 0.1,
) -> AssessModelFitResult:
    """Assess model fit for given diagnostics results.

    It determines if a model fit is good or bad based on Fisher exact test p

    Args:
        diagnostics: Output of compute_diagnostics

    Returns:
        Two dictionaries, one for good metrics, one for bad metrics, each
        mapping metric name to p-value
    """

    good_fit_metrics_to_fisher_score: dict[str, float] = {}
    bad_fit_metrics_to_fisher_score: dict[str, float] = {}

    for metric, score in diagnostics[FISHER_EXACT_TEST_P].items():
        if score > significance_level:
            bad_fit_metrics_to_fisher_score[metric] = score
        else:
            good_fit_metrics_to_fisher_score[metric] = score
    if len(bad_fit_metrics_to_fisher_score) > 0:
        logger.warning(
            "{} {} {} unable to be reliably fit.".format(
                ("Metrics" if len(bad_fit_metrics_to_fisher_score) > 1 else "Metric"),
                (" , ".join(bad_fit_metrics_to_fisher_score.keys())),
                ("were" if len(bad_fit_metrics_to_fisher_score) > 1 else "was"),
            )
        )
    return AssessModelFitResult(
        good_fit_metrics_to_fisher_score=good_fit_metrics_to_fisher_score,
        bad_fit_metrics_to_fisher_score=bad_fit_metrics_to_fisher_score,
    )


def has_good_opt_config_model_fit(
    optimization_config: OptimizationConfig,
    assess_model_fit_result: AssessModelFitResult,
) -> bool:
    """Assess model fit for given diagnostics results across the optimization
    config metrics

    Bad fit criteria: Any objective metrics are poorly fit based on
    the Fisher exact test p (see assess_model_fit())

    TODO[]: Incl. outcome constraints in assessment

    Args:
        optimization_config: Objective/Outcome constraint metrics to assess
        diagnostics: Output of compute_diagnostics

    Returns:
        Two dictionaries, one for good metrics, one for bad metrics, each
        mapping metric name to p-value
    """

    # Bad fit criteria: Any objective metrics are poorly fit
    # TODO[]: Incl. outcome constraints in assessment
    has_good_opt_config_fit = all(
        (m.name in assess_model_fit_result.good_fit_metrics_to_fisher_score)
        for m in optimization_config.objective.metrics
    )
    return has_good_opt_config_fit


def _kfold_train_test_split(
    folds: int,
    training_data: ExperimentData,
) -> Iterable[CVData]:
    """Return train/test CV splits based on arm names.

    Args:
        folds: Number of folds to return
        training_data: Training data to split

    Returns:
        Yields CVData object of train/test data.
    """
    arm_name_vals = set(training_data.arm_data.index.unique(level="arm_name"))
    n = len(arm_name_vals)
    if n < 2:
        raise UnsupportedError(
            "Cross validation requires at least two in-design arms in the training "
            f"data. Only {n} in-design arms were found."
        )
    elif folds > n:
        raise ValueError(
            f"Training data only has {n} arms, which is less than {folds} folds."
        )
    elif folds < 2 and folds != -1:
        raise ValueError("Folds must be -1 for LOO, or > 1.")
    elif folds == -1:
        folds = n

    arm_names = np.array(list(arm_name_vals))
    # Not necessary to shuffle when using LOO, avoids differences in floating point
    # computations making equality tests brittle.
    if folds != n:
        np.random.shuffle(arm_names)
    test_size = n // folds  # The size of all test sets but the last
    final_size = test_size + (n - folds * test_size)  # Grab the leftovers
    for fold in range(folds):
        # We will take the test set from the back of the array.
        # Roll the list of arm names to get a fresh test set
        arm_names = np.roll(arm_names, test_size)
        n_test = test_size if fold < folds - 1 else final_size
        train_names = set(arm_names[:-n_test])
        test_names = set(arm_names[-n_test:])
        yield CVData(
            training_data=training_data.filter_by_arm_names(arm_names=train_names),
            test_data=training_data.filter_by_arm_names(arm_names=test_names),
        )


def gen_trial_split(
    training_data: ExperimentData,
    test_trials: list[int],
    train_trials: list[int] | None = None,
) -> Iterable[CVData]:
    """Return a single train/test CV split based on trial index.

    Args:
        training_data: Training data to split
        test_trials: List of trial indices to use as test data

    Returns:
        A single CVData object of train/test data.
    """
    if len(test_trials) == 0:
        raise ValueError("No test trials provided.")
    all_trials = training_data.arm_data.index.get_level_values("trial_index")
    if set(test_trials) - set(all_trials):
        raise ValueError(
            f"Trials {test_trials} not all in training data trials {all_trials}."
        )
    if train_trials is None:
        train_trials = list(set(all_trials) - set(test_trials))
    else:
        if set(train_trials).intersection(set(test_trials)):
            raise ValueError("Test and train trials overlap.")
    if len(train_trials) == 0:
        raise ValueError(f"All trials in data, {all_trials}, specified as test trials.")
    logger.debug(f"Using trials {train_trials} for training.")
    train_data = training_data.filter_by_trial_index(trial_indices=train_trials)
    test_data = training_data.filter_by_trial_index(trial_indices=test_trials)
    return [CVData(training_data=train_data, test_data=test_data)]


"""
############################## Model Fit Metrics Utils ##############################
"""


def get_fit_and_std_quality_and_generalization_dict(
    fitted_adapter: Adapter,
) -> dict[str, float | None]:
    """
    Get stats and gen from a fitted Adapter for analytics purposes.
    """
    try:
        model_fit_dict = compute_model_fit_metrics_from_adapter(
            adapter=fitted_adapter,
            generalization=False,
            untransform=False,
        )
        # similar for uncertainty quantification, but distance from 1 matters
        std = list(model_fit_dict["std_of_the_standardized_error"].values())

        # generalization metrics
        model_gen_dict = compute_model_fit_metrics_from_adapter(
            adapter=fitted_adapter,
            generalization=True,
            untransform=False,
        )
        gen_std = list(model_gen_dict["std_of_the_standardized_error"].values())
        return {
            "model_fit_quality": _model_fit_metric(model_fit_dict),
            "model_std_quality": _model_std_quality(np.array(std)),
            "model_fit_generalization": _model_fit_metric(model_gen_dict),
            "model_std_generalization": _model_std_quality(np.array(gen_std)),
        }

    # Do not warn if the Adapter does not implement a predict method
    # (ex. RandomAdapter).
    except NotImplementedError:
        return {
            "model_fit_quality": None,
            "model_std_quality": None,
            "model_fit_generalization": None,
            "model_std_generalization": None,
        }

    except Exception as e:
        warn("Encountered exception in computing model fit quality: " + str(e))
        return {
            "model_fit_quality": None,
            "model_std_quality": None,
            "model_fit_generalization": None,
            "model_std_generalization": None,
        }


def compute_model_fit_metrics_from_adapter(
    adapter: Adapter,
    fit_metrics_dict: dict[str, ModelFitMetricProtocol] | None = None,
    generalization: bool = False,
    untransform: bool = False,
) -> dict[str, dict[str, float]]:
    """Computes the model fit metrics given a Adapter and an Experiment.

    Args:
        adapter: The Adapter for which to compute the model fit metrics.
        experiment: The experiment with whose data to compute the metrics if
            generalization == False. Otherwise, the data is taken from the Adapter.
        fit_metrics_dict: An optional dictionary with model fit metric functions,
            i.e. a ModelFitMetricProtocol, as values and their names as keys.
        generalization: Boolean indicating whether to compute the generalization
            metrics on cross-validation data or on the training data. The latter
            helps diagnose problems with model training, rather than generalization.
        untransform: Boolean indicating whether to untransform model predictions
            before calcualting the model fit metrics. False by default as models
            are trained in transformed space and model fit should be
            evaluated in transformed space.

    Returns:
        A nested dictionary mapping from the *model fit* metric names and the
        *experimental metric* names to the values of the model fit metrics.

        Example for an imaginary AutoML experiment that seeks to minimize the test
        error after training an expensive model, with respect to hyper-parameters:

        ```
        model_fit_dict = compute_model_fit_metrics_from_adapter(adapter, exp)
        model_fit_dict["coefficient_of_determination"]["test error"] =
            `coefficient of determination of the test error predictions`
        ```
    """
    predict_func = (
        _predict_on_cross_validation_data
        if generalization
        else _predict_on_training_data
    )
    y_obs, y_pred, se_pred = predict_func(adapter=adapter, untransform=untransform)
    if fit_metrics_dict is None:
        fit_metrics_dict = {
            "coefficient_of_determination": coefficient_of_determination,
            "mean_of_the_standardized_error": mean_of_the_standardized_error,
            "std_of_the_standardized_error": std_of_the_standardized_error,
        }

    return compute_model_fit_metrics(
        y_obs=y_obs,
        y_pred=y_pred,
        se_pred=se_pred,
        fit_metrics_dict=fit_metrics_dict,
    )


def _model_fit_metric(metric_dict: dict[str, dict[str, float]]) -> float:
    # We'd ideally log the entire `model_fit_dict` as a single model fit metric
    # can't capture the nuances of multiple experimental metrics, but this might
    # lead to database performance issues. So instead, we take the worst
    # coefficient of determination as model fit quality and store the full data
    # in Manifold (TODO).
    return min(metric_dict["coefficient_of_determination"].values())


def _model_std_quality(std: npt.NDArray) -> float:
    """Quantifies quality of the model uncertainty. A value of one means the
    uncertainty is perfectly predictive of the true standard deviation of the error.
    Values larger than one indicate over-estimation and negative values indicate
    under-estimation of the true standard deviation of the error. In particular, a value
    of 2 (resp. 1 / 2) represents an over-estimation (resp. under-estimation) of the
    true standard deviation of the error by a factor of 2.

    Args:
        std: The standard deviation of the standardized error.

    Returns:
        The factor corresponding to the worst over- or under-estimation factor of the
        standard deviation of the error among all experimentally observed metrics.
    """
    max_std, min_std = np.max(std), np.min(std)
    # comparing worst over-estimation factor with worst under-estimation factor
    inv_model_std_quality = max_std if max_std > 1 / min_std else min_std
    # reciprocal so that values greater than one indicate over-estimation and
    # values smaller than indicate underestimation of the uncertainty.
    return 1 / inv_model_std_quality


def _predict_on_training_data(
    adapter: Adapter,
    untransform: bool = False,
) -> tuple[
    dict[str, npt.NDArray],
    dict[str, npt.NDArray],
    dict[str, npt.NDArray],
]:
    """Makes predictions on the training data of a given experiment using a Adapter
    and returning the observed values, and the corresponding predictive means and
    predictive standard deviations of the model, in transformed space.

    NOTE: This is a helper function for `compute_model_fit_metrics_from_adapter`.

    Args:
        adapter: A Adapter object with which to make predictions.
        untransform: Boolean indicating whether to untransform model predictions.

    Returns:
        A tuple containing three dictionaries for 1) observed metric values, and the
        model's associated 2) predictive means and 3) predictive standard deviations.
    """
    training_data = adapter.get_training_data()
    observation_features = [
        ObservationFeatures(
            # NOTE: It is crucial to pop metadata first here.
            # Otherwise, it'd end up in parameters.
            metadata=row.pop("metadata"),
            parameters=row.to_dict(),
            trial_index=assert_is_instance(index, tuple)[0],
        )
        for index, row in training_data.arm_data.iterrows()
    ]
    observation_data_pred = adapter._predict_observation_data(
        observation_features=observation_features,
        untransform=untransform,
    )

    mean_predicted, cov_predicted = unwrap_observation_data(observation_data_pred)
    mean_observed = {
        name: col.to_numpy()
        for name, col in training_data.observation_data["mean"].items()
    }
    # Converting dictionary values to arrays.
    mean_predicted = {k: np.array(v) for k, v in mean_predicted.items()}
    std_predicted = {m: np.sqrt(np.array(cov_predicted[m][m])) for m in cov_predicted}
    return mean_observed, mean_predicted, std_predicted


def _predict_on_cross_validation_data(
    adapter: Adapter,
    untransform: bool = False,
) -> tuple[
    dict[str, npt.NDArray],
    dict[str, npt.NDArray],
    dict[str, npt.NDArray],
]:
    """Makes leave-one-out cross-validation predictions on the training data of the
    Adapter and returns the observed values, and the corresponding predictive means
    and predictive standard deviations of the model as numpy arrays,
    in transformed space.

    NOTE: This is a helper function for `compute_model_fit_metrics_from_adapter`.

    Args:
        adapter: A Adapter object with which to make predictions.
        untransform: Boolean indicating whether to untransform model predictions
            before cross validating. False by default.

    Returns:
        A tuple containing three dictionaries, each mapping metric_name to:
            1. observed metric values,
            2. LOOCV predicted mean at each observed point, and
            3. LOOCV predicted standard deviation at each observed point.
    """
    cv = cross_validate(model=adapter, untransform=untransform)

    metric_names = cv[0].observed.data.metric_names
    mean_observed = {k: [] for k in metric_names}
    mean_predicted = {k: [] for k in metric_names}
    std_predicted = {k: [] for k in metric_names}

    for cvi in cv:
        obs = cvi.observed.data
        for k, v in zip(obs.metric_names, obs.means):
            mean_observed[k].append(v)

        pred = cvi.predicted
        for k, v in zip(pred.metric_names, pred.means):
            mean_predicted[k].append(v)

        pred_se = np.sqrt(pred.covariance.diagonal().clip(0))
        for k, v in zip(pred.metric_names, pred_se):
            std_predicted[k].append(v)

    mean_observed = {k: np.array(v) for k, v in mean_observed.items()}
    mean_predicted = {k: np.array(v) for k, v in mean_predicted.items()}
    std_predicted = {k: np.array(v) for k, v in std_predicted.items()}
    return mean_observed, mean_predicted, std_predicted



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable, Iterable, Mapping, Sequence

from dataclasses import dataclass, field, InitVar
from functools import partial
from logging import Logger
from typing import Any, Protocol, TypeVar

import numpy as np
import numpy.typing as npt
import pandas as pd
import torch
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.benchmark_test_function import BenchmarkTestFunction
from ax.benchmark.problems.surrogate.lcbench.data import (
    DATASET_NAMES,
    LCBenchData,
    load_lcbench_data,
)
from ax.benchmark.problems.surrogate.lcbench.transfer_learning import DEFAULT_NUM_TRIALS
from ax.benchmark.problems.surrogate.lcbench.utils import (
    BASELINE_VALUES,
    DEFAULT_METRIC_NAME,
    get_lcbench_log_scale_parameter_names,
    get_lcbench_optimization_config,
    get_lcbench_parameter_names,
    get_lcbench_search_space,
)
from ax.core.optimization_config import OptimizationConfig
from ax.core.search_space import SearchSpace
from ax.core.types import TParamValue
from ax.exceptions.core import UserInputError
from ax.utils.common.logger import get_logger

from sklearn.compose import make_column_transformer, TransformedTargetRegressor
from sklearn.ensemble import RandomForestRegressor
from sklearn.pipeline import make_pipeline
from sklearn.preprocessing import FunctionTransformer, MinMaxScaler


logger: Logger = get_logger(__name__)

TRegressorProtocol = TypeVar("TRegressorProtocol", bound="RegressorProtocol")


OPTIMAL_VALUES: dict[str, float] = {
    "APSFailure": 98.97643280029295,
    "Amazon_employee_access": 94.1208953857422,
    "Australian": 92.1568603515625,
    "Fashion-MNIST": 89.417236328125,
    "KDDCup09_appetency": 98.33574676513672,
    "MiniBooNE": 90.79180908203124,
    "adult": 84.11111450195312,
    "airlines": 63.999061584472656,
    "albert": 66.25859832763672,
    "bank-marketing": 88.8466567993164,
    "blood-transfusion-service-center": 80.1204833984375,
    "car": 96.8668441772461,
    "christine": 75.04173278808594,
    "cnae-9": 97.48954010009766,
    "connect-4": 74.20671844482422,
    "covertype": 77.28044128417969,
    "credit-g": 77.47747802734375,
    "dionis": 87.68419647216797,
    "fabert": 67.32564544677734,
    "helena": 29.10163116455078,
    "higgs": 71.88652801513672,
    "jannis": 66.58744812011719,
    "jasmine": 82.60211944580078,
    "jungle_chess_2pcs_raw_endgame_complete": 84.13723754882812,
    "kc1": 87.36616516113281,
    "kr-vs-kp": 99.00990295410156,
    "mfeat-factors": 97.96839904785156,
    "nomao": 96.2865753173828,
    "numerai28.6": 52.32661819458008,
    "phoneme": 82.09204864501953,
    "segment": 90.60665130615234,
    "shuttle": 99.76605987548828,
    "sylvine": 94.97354125976562,
    "vehicle": 81.9148941040039,
    "volkert": 64.02699279785156,
}

# Chosen so that for the median parameterization, one step takes one virtual
# second.
RUNTIME_MULTIPLIERS = {
    "APSFailure": 0.2498361853616576,
    "Amazon_employee_access": 0.5596248586092226,
    "Australian": 1.0748285031033658,
    "Fashion-MNIST": 0.057873893027107395,
    "KDDCup09_appetency": 0.2714829300383819,
    "MiniBooNE": 0.20530997463252754,
    "adult": 0.4423476551967684,
    "airlines": 0.06473793535537586,
    "albert": 0.06832237841522835,
    "bank-marketing": 0.46517394252532845,
    "blood-transfusion-service-center": 1.1101296769694071,
    "car": 1.0536256049024968,
    "christine": 0.025742718302424954,
    "cnae-9": 0.08811760797353926,
    "connect-4": 0.33489219890695243,
    "covertype": 0.05049155246078877,
    "credit-g": 1.0400726314123157,
    "dionis": 0.0231601276801126,
    "fabert": 0.08971358669025394,
    "helena": 0.25008050673472376,
    "higgs": 0.26990484596881176,
    "jannis": 0.2828372999943685,
    "jasmine": 0.7655180467265444,
    "jungle_chess_2pcs_raw_endgame_complete": 0.47160243094434906,
    "kc1": 1.0143178289557349,
    "kr-vs-kp": 0.9390239320512418,
    "mfeat-factors": 0.595676967891612,
    "nomao": 0.4420599860962263,
    "numerai28.6": 0.28377545234818863,
    "phoneme": 0.9689051773179346,
    "segment": 1.000676324600838,
    "shuttle": 0.39362569776573014,
    "sylvine": 0.9179851039769921,
    "vehicle": 1.048848701347826,
    "volkert": 0.28538440509808005,
}


class RegressorProtocol(Protocol):
    """
    A regressor that can fit and predict, such as `RandomForestRegressor`.
    """

    def fit(
        self: TRegressorProtocol, X: pd.DataFrame, y: pd.DataFrame | pd.Series
    ) -> TRegressorProtocol: ...
    def predict(self: TRegressorProtocol, X: pd.DataFrame) -> npt.NDArray: ...
    def set_params(self: TRegressorProtocol, **kwargs: Any) -> TRegressorProtocol: ...


def get_default_base_regressor() -> RegressorProtocol:
    return RandomForestRegressor(max_depth=30)


def _create_surrogate_regressor(
    base_regressor: RegressorProtocol,
    log_numeric_columns: Iterable,
    numeric_columns: Iterable,
    seed: int = 0,
) -> RegressorProtocol:
    unit_scaler = MinMaxScaler()
    log_transformer = FunctionTransformer(
        func=np.log, inverse_func=np.exp, validate=True
    )

    log_numeric_transformer = make_pipeline(log_transformer, unit_scaler)

    preprocessor = make_column_transformer(
        (log_numeric_transformer, list(log_numeric_columns)),
        (unit_scaler, list(numeric_columns)),
        remainder="drop",
    )

    try:
        regressor = base_regressor.set_params(random_state=seed)
    except ValueError:
        # some models (e.g. K nearest neighbors) are deterministic by nature and do not
        # allow you to set a random seed
        logger.warning(
            f"Surrogate model `{base_regressor}` does not support specification of "
            "random seed, which *may* indicate that the model is already "
            "deterministic by nature. However, if you're unsure, this could lead to "
            "non-deterministic behavior in your experiments."
        )
        regressor = base_regressor

    return make_pipeline(
        preprocessor,
        TransformedTargetRegressor(regressor=regressor, transformer=log_transformer),
    )


@dataclass(kw_only=True)
class LearningCurveBenchmarkTestFunction(BenchmarkTestFunction):
    """A benchmark test function for LCBench early-stopping problems.

    This class represents a learning curve benchmark test function, which leverages a
    surrogate model trained to predict the performance of deep learning models at
    different stages of training. The test function takes in a set of hyperparameters
    and returns a tensor representing the predicted performance of the model at each
    stage (epoch) of training.

    To use this class, you would typically create an instance of it and pass it to a
    `BenchmarkProblem` along with a search space, optimization config, and other
    relevant parameters.

    Example:
        test_function = LearningCurveBenchmarkTestFunction(dataset_name="vehicle")
        search_space = get_lcbench_search_space()
        optimization_config = get_lcbench_optimization_config(
            metric_name="Train/val_accuracy", observe_noise_sd=True, use_map_metric=True
        )
        problem = BenchmarkProblem(
            name="vehicle_Train/val_accuracy",
            search_space=search_space,
            optimization_config=optimization_config,
            test_function=test_function,
            step_runtime_function=None,
            ...
        )
    """

    n_steps: int = field(init=False)
    outcome_names: Sequence[str] = field(default_factory=lambda: [DEFAULT_METRIC_NAME])
    dataset_name: str
    metric_surrogate: RegressorProtocol = field(init=False)
    runtime_surrogate: RegressorProtocol = field(init=False)

    # pyre-ignore [16]: Pyre doesn't understand InitVars.
    metric_base_surrogate: InitVar[RegressorProtocol] = get_default_base_regressor()
    # pyre-ignore [16]: Pyre doesn't understand InitVars.
    runtime_base_surrogate: InitVar[RegressorProtocol] = get_default_base_regressor()

    def __post_init__(
        self,
        metric_base_surrogate: RegressorProtocol,
        runtime_base_surrogate: RegressorProtocol,
    ) -> None:
        if len(self.outcome_names) != 1:  # pragma: no cover
            raise ValueError("Exactly one outcome is supported currently")

        metric_name = self.outcome_names[0]
        lcbench_data = self._load_lcbench_data(metric_name)
        self.n_steps = lcbench_data.metric_df.shape[-1]

        parameter_names = get_lcbench_parameter_names()
        log_scale_parameter_names = get_lcbench_log_scale_parameter_names()
        numeric_columns = set(parameter_names) - set(log_scale_parameter_names)

        self.metric_surrogate = _create_surrogate_regressor(
            base_regressor=metric_base_surrogate,
            log_numeric_columns=log_scale_parameter_names,
            numeric_columns=numeric_columns,
        ).fit(X=lcbench_data.parameter_df, y=lcbench_data.metric_df)
        self.runtime_surrogate = _create_surrogate_regressor(
            base_regressor=runtime_base_surrogate,
            log_numeric_columns=log_scale_parameter_names,
            numeric_columns=numeric_columns,
        ).fit(X=lcbench_data.parameter_df, y=lcbench_data.average_runtime_series)

    def _load_lcbench_data(self, metric_name: str) -> LCBenchData:
        return load_lcbench_data(
            dataset_name=self.dataset_name,
            metric_name=metric_name,
            log_scale_parameter_names=[],
        )

    def evaluate_true(self, params: Mapping[str, TParamValue]) -> torch.Tensor:
        X = pd.DataFrame.from_records(data=[params])
        Y = self.metric_surrogate.predict(X=X)  # shape: (1, 50)
        return torch.from_numpy(Y)

    def step_runtime(self, params: Mapping[str, TParamValue]) -> float:
        X = pd.DataFrame.from_records(data=[params])
        Y = self.runtime_surrogate.predict(X=X)  # shape: (1,)
        return Y.item() * RUNTIME_MULTIPLIERS[self.dataset_name]


def _get_lcbench_early_stopping_benchmark_problem(
    dataset_name: str,
    test_function_class: type[LearningCurveBenchmarkTestFunction],
    num_trials: int = DEFAULT_NUM_TRIALS,
    constant_step_runtime: bool = False,
) -> BenchmarkProblem:
    """Construct an LCBench early-stopping benchmark problem.

    Args:
        dataset_name: Must be one of the keys of `DEFAULT_AND_OPTIMAL_VALUES`, which
            correspond to the names of the datasets available in LCBench.
        num_trials: The number of optimization trials to run.
        constant_step_runtime: Determines if the step runtime is fixed or varies
            based on the hyperparameters.

    Returns:
        An LCBench surrogate benchmark problem.
    """

    if dataset_name not in DATASET_NAMES:
        raise UserInputError(f"`dataset_name` must be one of {sorted(DATASET_NAMES)}")

    optimal_value = OPTIMAL_VALUES[dataset_name]
    baseline_value = BASELINE_VALUES[dataset_name]

    search_space: SearchSpace = get_lcbench_search_space()
    optimization_config: OptimizationConfig = get_lcbench_optimization_config(
        metric_name=DEFAULT_METRIC_NAME,
        observe_noise_sd=False,
        use_map_metric=True,
    )

    test_function = test_function_class(dataset_name=dataset_name)
    step_runtime_function = (
        None if constant_step_runtime else test_function.step_runtime
    )

    return BenchmarkProblem(
        name=f"LCBench::{dataset_name}",
        search_space=search_space,
        optimization_config=optimization_config,
        num_trials=num_trials,
        optimal_value=optimal_value,
        baseline_value=baseline_value,
        test_function=test_function,
        step_runtime_function=step_runtime_function,
    )


get_lcbench_early_stopping_benchmark_problem: Callable[..., BenchmarkProblem] = partial(
    _get_lcbench_early_stopping_benchmark_problem,
    test_function_class=LearningCurveBenchmarkTestFunction,
)



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from collections.abc import Callable
from logging import Logger
from typing import Any, Union

import pandas as pd
from ax.adapter import Adapter
from ax.analysis.analysis import Analysis
from ax.analysis.healthcheck.healthcheck_analysis import (
    create_healthcheck_analysis_card,
    HealthcheckAnalysisCard,
    HealthcheckStatus,
)
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.logger import get_logger
from pyre_extensions import override

logger: Logger = get_logger(__name__)


class MetricFetchingErrorsAnalysis(Analysis):
    """
    Analysis to check if any metric fetch errors occurred.
    If any metric fetch errors have occurred, the analysis will display the
    trial index, metric name, timestamp, error message, and traceback.
    """

    def __init__(
        self,
        add_traceback_paste_callable: Callable[
            [str, list[dict[str, Any]]], list[dict[str, Any]]
        ]
        | None = None,
    ) -> None:
        r"""
        Args:
            max_records: Max number of records to display.
            add_traceback_paste_callable: A function that takes an experiment name and a
            list of metric fetch errors, and returns the modified list with a paste
            link added to each error's traceback if available. If no function is
            provided, no traceback information will be included in the health check.
        """
        self.add_traceback_paste_callable = add_traceback_paste_callable

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> HealthcheckAnalysisCard:
        if experiment is None:
            raise UserInputError("MetricFetchingErrorsAnalysis requires an Experiment.")

        df = pd.DataFrame()

        metric_fetch_errors = experiment._metric_fetching_errors

        if len(metric_fetch_errors) == 0:
            return create_healthcheck_analysis_card(
                name=self.__class__.__name__,
                title="Metric Fetch Errors",
                subtitle="No metric fetch errors found.",
                df=df,
                status=HealthcheckStatus.PASS,
            )

        metric_fetch_errors = [
            e for e in metric_fetch_errors.values() if self._validate_fields(errors=e)
        ]

        metric_fetch_errors = sorted(
            metric_fetch_errors, key=lambda e: e["timestamp"], reverse=True
        )

        if self.add_traceback_paste_callable is not None:
            metric_fetch_errors = self.add_traceback_paste_callable(
                experiment.name, metric_fetch_errors
            )
        else:
            for e in metric_fetch_errors:
                e["paste"] = "No traceback available"

        metric_fetch_errors_for_card = [
            {
                "trial_index": e["trial_index"],
                "metric_name": e["metric_name"],
                "timestamp": e["timestamp"],
                "reason": e["reason"],
                "traceback": e["paste"],
            }
            for e in metric_fetch_errors
        ]

        df = pd.DataFrame(
            columns=[
                "trial_index",
                "metric_name",
                "timestamp",
                "reason",
                "traceback",
            ]
        )
        df = pd.concat([df, pd.DataFrame.from_records(metric_fetch_errors_for_card)])

        subtitle_df_columns = {
            "trial_index": "Trial Index",
            "metric_name": "Metric Name",
            "timestamp": "Timestamp",
            "reason": "Error Message",
            "traceback": "Traceback",
        }
        subtitle = df.rename(columns=subtitle_df_columns)

        return create_healthcheck_analysis_card(
            name=self.__class__.__name__,
            title="Metric Fetch Errors",
            subtitle=subtitle.to_markdown(index=False),
            df=df,
            status=HealthcheckStatus.WARNING,
        )

    @staticmethod
    def _validate_fields(errors: dict[str, Union[int, str]]) -> bool:
        required_fields = {
            "trial_index",
            "metric_name",
            "reason",
            "timestamp",
            "traceback",
        }

        if not required_fields.issubset(errors.keys()):
            missing_fields = required_fields - errors.keys()
            logger.warning(
                f"Metric fetch error {errors} is missing required fields "
                f"{missing_fields}. Discarding."
            )
            return False

        # maybe it was already modified to have a paste since dicts are
        # passed by reference
        optional_fields = {"paste"}
        if not set(errors.keys()).issubset(required_fields.union(optional_fields)):
            extra_fields = set(errors.keys()) - required_fields.union(optional_fields)
            logger.warning(
                f"Metric fetch error {errors} has extra fields "
                f"{extra_fields}. Discarding."
            )
            return False

        return True



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from itertools import product

import numpy as np
import pandas as pd
from ax.analysis.plotly.utils import truncate_label
from ax.analysis.utils import _relativize_df_with_sq, prepare_arm_data
from ax.api.client import Client
from ax.api.configs import RangeParameterConfig
from ax.core.arm import Arm
from ax.core.batch_trial import BatchTrial
from ax.core.experiment import Experiment
from ax.core.metric import Metric
from ax.core.trial_status import TrialStatus  # noqa
from ax.exceptions.core import UserInputError
from ax.utils.common.testutils import TestCase
from ax.utils.stats.statstools import relativize_data
from ax.utils.testing.core_stubs import get_offline_experiments, get_online_experiments
from ax.utils.testing.mock import mock_botorch_optimize
from ax.utils.testing.modeling_stubs import get_default_generation_strategy_at_MBM_node
from pyre_extensions import assert_is_instance, none_throws


class TestUtils(TestCase):
    @mock_botorch_optimize
    def setUp(self) -> None:
        super().setUp()

        self.client = Client()
        self.client.configure_experiment(
            name="test_experiment",
            parameters=[
                RangeParameterConfig(
                    name="x1",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
                RangeParameterConfig(
                    name="x2",
                    parameter_type="float",
                    bounds=(0, 1),
                ),
            ],
        )
        self.client.configure_optimization(
            objective="foo",
            outcome_constraints=["bar >= 11", "baz <= 18", "qux >= 1998"],
        )

        # Get two trials and fail one, giving us a ragged structure
        self.client.get_next_trials(max_trials=2)
        sem = 0.1
        self.client.complete_trial(
            trial_index=0,
            raw_data={
                "foo": (1.0, sem),
                "bar": (2.0, sem),
                "baz": (3.0, sem),
                "qux": (4.0, sem),
            },
        )
        self.client.mark_trial_failed(trial_index=1)

        # Complete 5 trials successfully
        for _ in range(5):
            for trial_index, parameterization in self.client.get_next_trials(
                max_trials=1
            ).items():
                self.client.complete_trial(
                    trial_index=trial_index,
                    raw_data={
                        "foo": (assert_is_instance(parameterization["x1"], float), sem),
                        "bar": (
                            assert_is_instance(parameterization["x1"], float)
                            - 2 * assert_is_instance(parameterization["x2"], float),
                            sem,
                        ),
                        "baz": (3.0, sem),
                        "qux": (4.0, sem),
                    },
                )

    def test_prepare_arm_data_validation(self) -> None:
        with self.assertRaisesRegex(
            UserInputError, "Must provide at least one metric name"
        ):
            prepare_arm_data(
                experiment=self.client._experiment,
                metric_names=[],
                use_model_predictions=False,
            )

        with self.assertRaisesRegex(
            UserInputError, "Requested metrics .* are not present in the experiment."
        ):
            prepare_arm_data(
                experiment=self.client._experiment,
                metric_names=["foo", "bar", "zed"],
                use_model_predictions=False,
            )

        with self.assertRaisesRegex(
            UserInputError, "Trial with index .* not found in experiment."
        ):
            prepare_arm_data(
                experiment=self.client._experiment,
                metric_names=["foo", "bar"],
                trial_index=1998,
                use_model_predictions=False,
            )

        with self.assertRaisesRegex(UserInputError, "Must provide an adapter"):
            prepare_arm_data(
                experiment=self.client._experiment,
                metric_names=["foo", "bar"],
                use_model_predictions=True,
            )

        with self.assertRaisesRegex(
            UserInputError,
            "Cannot provide additional arms when use_model_predictions=False.",
        ):
            prepare_arm_data(
                experiment=self.client._experiment,
                metric_names=["foo", "bar"],
                use_model_predictions=False,
                additional_arms=[Arm(parameters={"x1": 0.5, "x2": 0.5})],
            )

    def test_prepare_arm_data_raw(self) -> None:
        df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar", "baz", "qux"],
            use_model_predictions=False,
        )

        # Check that the columns are correct
        self.assertEqual(
            set(df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
                "baz_mean",
                "baz_sem",
                "qux_mean",
                "qux_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(len(df), len(self.client._experiment.arms_by_name))
        for arm_name in self.client._experiment.arms_by_name:
            self.assertEqual((df["arm_name"] == arm_name).sum(), 1)

        # Check that the FAILED trial has no mean or sem
        self.assertTrue(np.isnan(df.loc[1]["foo_mean"]))
        self.assertTrue(np.isnan(df.loc[1]["foo_sem"]))

        # Check that all SEMs have the correct value
        success_idcs = [i for i in range(df.shape[0]) if i != 1]
        self.assertTrue((df.loc[success_idcs]["foo_sem"] == 0.1).all())
        self.assertTrue((df.loc[success_idcs]["bar_sem"] == 0.1).all())

        # Check that p_feasible_mean is NaN for the arm without data and not NaN for the
        # other arms.
        self.assertTrue(np.isnan(df.loc[1]["p_feasible_mean"]))
        self.assertFalse(df[df["arm_name"] != "1_0"]["p_feasible_mean"].isna().any())
        self.assertTrue(df["p_feasible_sem"].isna().all())

        only_foo_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo"],
            use_model_predictions=False,
        )
        # Check that the columns are correct
        self.assertEqual(
            set(only_foo_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(len(only_foo_df), len(self.client._experiment.arms_by_name))
        for arm_name in self.client._experiment.arms_by_name:
            self.assertEqual((only_foo_df["arm_name"] == arm_name).sum(), 1)

        # Check that all SEMs have the correct value
        self.assertTrue((only_foo_df.loc[success_idcs]["foo_sem"] == 0.1).all())

        only_trial_0_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=False,
            trial_index=0,
        )

        # Check that the columns are correct
        self.assertEqual(
            set(only_trial_0_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        # Check that we have one row per arm and that the arm appears only once
        self.assertEqual(len(only_trial_0_df), 1)
        self.assertEqual(only_trial_0_df.loc[0]["arm_name"], "0_0")

        # Check that all means are not NaN
        self.assertFalse(only_trial_0_df["foo_mean"].isna().any())

        # Check that all SEMs have the correct value
        self.assertTrue((only_trial_0_df["foo_sem"] == 0.1).all())

        only_completed_trials_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=False,
            trial_statuses=[TrialStatus.COMPLETED],
        )

        # Check that we have two rows per arm and that each arm appears only once
        self.assertEqual(
            len(only_completed_trials_df), len(self.client._experiment.arms_by_name) - 1
        )

        # Check that all means are not NaN
        self.assertFalse(only_completed_trials_df["foo_mean"].isna().any())

        # Check that all SEMs have the correct value
        self.assertTrue((only_completed_trials_df["foo_sem"] == 0.1).all())

    def test_prepare_arm_data_use_model_predictions_with_abandoned_arms(self) -> None:
        search_space = self.client._experiment.search_space.clone()

        # Configure an experiment with batch trials
        test_exp = Experiment(
            name="test_exp_multifeed",
            search_space=search_space,
            description="test MF experiment",
            tracking_metrics=[
                Metric(name="foo"),
                Metric(name="bar"),
                Metric(name="baz"),
                Metric(name="qux"),
            ],
            is_test=True,
        )
        trial = BatchTrial(test_exp)

        # set 2 arms and abandon one
        trial.add_arms_and_weights(
            [
                Arm(name="0_0", parameters={"x1": 0.5, "x2": 0.5}),
                Arm(name="0_1", parameters={"x1": 1.5, "x2": 1.5}),
            ]
        )
        trial.mark_arm_abandoned(arm_name="0_1")
        trial.attach_batch_trial_data(
            raw_data={
                "0_0": {
                    "foo": 1.0,
                    "bar": 2.0,
                    "baz": 3.0,
                    "qux": 4.0,
                }
            }
        )

        df = prepare_arm_data(
            experiment=test_exp,
            metric_names=["foo", "bar", "baz", "qux"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
        )

        self.assertIn("0_0", df["arm_name"].values)
        self.assertNotIn("0_1", df["arm_name"].values)

    def test_prepare_arm_data_use_model_predictions(self) -> None:
        df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar", "baz", "qux"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
        )

        # Check that the columns are correct
        self.assertEqual(
            set(df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
                "baz_mean",
                "baz_sem",
                "qux_mean",
                "qux_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(len(df), len(self.client._experiment.arms_by_name))
        for arm_name in self.client._experiment.arms_by_name:
            self.assertEqual((df["arm_name"] == arm_name).sum(), 1)

        # Check that all means and SEMs are not NaN
        self.assertFalse(df["foo_mean"].isna().any())
        self.assertFalse(df["foo_sem"].isna().any())
        self.assertFalse(df["bar_mean"].isna().any())
        self.assertFalse(df["bar_sem"].isna().any())

        # Check that all p_feasible_mean are not NaN
        self.assertFalse(df["p_feasible_mean"].isna().any())
        # Check that all p_feasible_sem are not NaN
        self.assertTrue(df["p_feasible_sem"].isna().all())

        only_foo_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
        )
        # Check that the columns are correct
        self.assertEqual(
            set(only_foo_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(len(only_foo_df), len(self.client._experiment.arms_by_name))
        for arm_name in self.client._experiment.arms_by_name:
            self.assertEqual((only_foo_df["arm_name"] == arm_name).sum(), 1)

        # Check that all means and SEMs are not NaN
        self.assertFalse(only_foo_df["foo_mean"].isna().any())
        self.assertFalse(only_foo_df["foo_sem"].isna().any())

        only_trial_0_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
            trial_index=0,
        )

        # Check that the columns are correct
        self.assertEqual(
            set(only_trial_0_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        # Check that we have one row per arm and that the arm appears only once
        self.assertEqual(len(only_trial_0_df), 1)
        self.assertEqual(only_trial_0_df.loc[0]["arm_name"], "0_0")

        # Check that all means and SEMs are not NaN
        self.assertFalse(only_trial_0_df["foo_mean"].isna().any())
        self.assertFalse(only_trial_0_df["foo_sem"].isna().any())
        self.assertFalse(only_trial_0_df["bar_mean"].isna().any())
        self.assertFalse(only_trial_0_df["bar_sem"].isna().any())

        with_additional_arms_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
            additional_arms=[
                Arm(parameters={"x1": 0.5, "x2": 0.5}),
                Arm(parameters={"x1": 0.25, "x2": 0.25}),
            ],
        )

        # Check that the columns are correct
        self.assertEqual(
            set(with_additional_arms_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(
            len(with_additional_arms_df), len(self.client._experiment.arms_by_name) + 2
        )
        for arm_name in self.client._experiment.arms_by_name:
            self.assertEqual((with_additional_arms_df["arm_name"] == arm_name).sum(), 1)

        # Check that all means and SEMs are not NaN
        self.assertFalse(with_additional_arms_df["foo_mean"].isna().any())
        self.assertFalse(with_additional_arms_df["foo_sem"].isna().any())
        self.assertFalse(with_additional_arms_df["bar_mean"].isna().any())
        self.assertFalse(with_additional_arms_df["bar_sem"].isna().any())

        with_only_additional_arms_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
            trial_index=-1,
            additional_arms=[
                Arm(parameters={"x1": 0.5, "x2": 0.5}),
                Arm(parameters={"x1": 0.25, "x2": 0.25}),
            ],
        )

        # Check that the columns are correct
        self.assertEqual(
            set(with_only_additional_arms_df.columns),
            {
                "trial_index",
                "arm_name",
                "trial_status",
                "generation_node",
                "p_feasible_mean",
                "p_feasible_sem",
                "foo_mean",
                "foo_sem",
                "bar_mean",
                "bar_sem",
            },
        )

        # Check that we have one row per arm and that each arm appears only once
        self.assertEqual(len(with_only_additional_arms_df), 2)

        # Check that all means and SEMs are not NaN
        self.assertFalse(with_additional_arms_df["foo_mean"].isna().any())
        self.assertFalse(with_additional_arms_df["foo_sem"].isna().any())
        self.assertFalse(with_additional_arms_df["bar_mean"].isna().any())
        self.assertFalse(with_additional_arms_df["bar_sem"].isna().any())

        only_completed_trials_df = prepare_arm_data(
            experiment=self.client._experiment,
            metric_names=["foo", "bar"],
            use_model_predictions=True,
            adapter=self.client._generation_strategy.adapter,
            trial_statuses=[TrialStatus.COMPLETED],
        )

        # Check that we have two rows per arm and that each arm appears only once
        self.assertEqual(
            len(only_completed_trials_df), len(self.client._experiment.arms_by_name) - 1
        )

        # Check that all means are not NaN
        self.assertFalse(only_completed_trials_df["foo_mean"].isna().any())

        # Check that all SEMs are not NaN
        self.assertFalse(only_completed_trials_df["foo_sem"].isna().any())

    def test_relativize_df_with_sq(self) -> None:
        df = pd.DataFrame(
            {
                "trial_index": [0, 0, 0],
                "arm_name": ["status_quo", "arm1", "arm2"],
                "foo_mean": [10.0, 12.0, 15.0],
                "foo_sem": [1.0, 1.2, 1.5],
                "bar_mean": [20.0, 22.0, 25.0],
                "bar_sem": [2.0, 2.2, 2.5],
            }
        )

        rel_df = _relativize_df_with_sq(
            df=df, status_quo_df=df[df["arm_name"] == "status_quo"]
        )

        np.testing.assert_almost_equal(
            rel_df.loc[0, "foo_mean"], 0.0, decimal=1
        )  # status quo
        np.testing.assert_almost_equal(rel_df.loc[1, "foo_mean"], 0.2, decimal=1)
        np.testing.assert_almost_equal(rel_df.loc[2, "foo_mean"], 0.5, decimal=1)
        np.testing.assert_almost_equal(rel_df.loc[0, "foo_sem"], 0.1, decimal=1)
        np.testing.assert_almost_equal(rel_df.loc[1, "foo_sem"], 0.2, decimal=1)
        np.testing.assert_almost_equal(rel_df.loc[1, "bar_mean"], 0.1, decimal=1)
        np.testing.assert_almost_equal(rel_df.loc[1, "bar_sem"], 0.2, decimal=1)

    def test_relativize_df_with_sq_multiple_trials(self) -> None:
        df = pd.DataFrame(
            {
                "trial_index": [0, 0, 1, 1],
                "arm_name": ["status_quo", "arm1", "status_quo", "arm2"],
                "foo_mean": [10.0, 12.0, 10.0, 15.0],
                "foo_sem": [1.0, 1.2, 1.0, 1.5],
                "bar_mean": [20.0, 22.0, 20.0, 25.0],
                "bar_sem": [2.0, 2.2, 2.0, 2.5],
            }
        )

        rel_df = _relativize_df_with_sq(
            df=df, status_quo_df=df[df["arm_name"] == "status_quo"]
        )

        np.testing.assert_almost_equal(
            rel_df[(rel_df["trial_index"] == 0) & (rel_df["arm_name"] == "status_quo")][
                "foo_mean"
            ].iloc[0],
            0.0,
            decimal=1,
        )
        np.testing.assert_almost_equal(
            rel_df[(rel_df["trial_index"] == 0) & (rel_df["arm_name"] == "arm1")][
                "foo_mean"
            ].iloc[0],
            0.2,
            decimal=1,
        )
        np.testing.assert_almost_equal(
            rel_df[(rel_df["trial_index"] == 1) & (rel_df["arm_name"] == "status_quo")][
                "foo_mean"
            ].iloc[0],
            0.0,
            decimal=1,
        )
        np.testing.assert_almost_equal(
            rel_df[(rel_df["trial_index"] == 1) & (rel_df["arm_name"] == "arm2")][
                "foo_mean"
            ].iloc[0],
            0.5,
            decimal=1,
        )

    def test_truncate_label(self) -> None:
        with self.subTest("No truncation"):
            self.assertEqual(
                truncate_label(label="short:overall_END_ALL_CAP"),
                "short:overall_END_ALL_CAP",
            )
        with self.subTest("remove suffix"):
            self.assertEqual(
                truncate_label(label="not_as_short:overall_END_ALL_CAP", n=20),
                "not_as_short:overall",
            )
        with self.subTest("remove prefix"):
            self.assertEqual(
                truncate_label(
                    label="long_prefix:not_as_short:overall_END_ALL_CAP", n=20
                ),
                "not_as_short:overall",
            )
        with self.subTest("remove common words and keep two segments"):
            self.assertEqual(
                truncate_label(
                    label="long_prefix:short_prefix:not_as_short:overall_END_ALL_CAP",
                    n=30,
                ),
                "short_prefix:not_as_short",
            )
        with self.subTest("only one long label"):
            self.assertEqual(
                truncate_label(
                    label="thisismyextrabiglabelthatcantbehandled",
                    n=20,
                ),
                "thisismyextrabigl...",
            )

    @TestCase.ax_long_test(
        reason=(
            "Adapter.predict still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_online(self) -> None:
        # Test ArmEffectsPlot can be computed for a variety of experiments which
        # resemble those we see in an online setting.
        for experiment in get_online_experiments():
            data = experiment.lookup_data()
            rel_df = relativize_data(data).df
            raw_df = data.df
            metric_name = next(iter(experiment.metrics.keys()))
            raw_arm_value = raw_df[
                (raw_df.arm_name == "0_0") & (raw_df.metric_name == metric_name)
            ]["mean"].values[0]
            for (
                use_model_predictions,
                trial_index,
                with_additional_arms,
                relativize,
            ) in product(
                (True, False),
                (None, 0),
                (True, False),
                (True, False),
            ):
                if use_model_predictions and with_additional_arms:
                    additional_arms = [
                        Arm(
                            parameters={
                                parameter_name: 0
                                for parameter_name in (
                                    experiment.search_space.parameters.keys()
                                )
                            }
                        )
                    ]
                else:
                    additional_arms = None

                generation_strategy = get_default_generation_strategy_at_MBM_node(
                    experiment=experiment
                )
                generation_strategy.current_node._fit(experiment=experiment)
                adapter = none_throws(generation_strategy.adapter)
                df = prepare_arm_data(
                    experiment=experiment,
                    metric_names=[*adapter.metric_names],
                    use_model_predictions=use_model_predictions,
                    adapter=adapter,
                    trial_index=trial_index,
                    additional_arms=additional_arms,
                    relativize=relativize,
                )

                arm_df = df[df.arm_name == "0_0"]
                if relativize:
                    # test that data is relativized
                    expected_arm_value = rel_df[
                        (rel_df.arm_name == "0_0") & (rel_df.metric_name == metric_name)
                    ]["mean"].values[0]
                else:
                    # test that data is not relativized
                    expected_arm_value = raw_arm_value
                self.assertAllClose(
                    arm_df[f"{metric_name}_mean"].values[0],
                    expected_arm_value,
                    rtol=1e-4 if use_model_predictions else 1e-5,
                    atol=1e-3 if use_model_predictions else 1e-8,
                )

    @TestCase.ax_long_test(
        reason=(
            "Adapter.predict still too slow under @mock_botorch_optimize for this test"
        )
    )
    @mock_botorch_optimize
    def test_offline(self) -> None:
        # Test compute_arm_data can be computed for a variety of experiments which
        # resemble those we see in an offline setting.

        for experiment in get_offline_experiments():
            for use_model_predictions in [True, False]:
                for trial_index in [None, 0]:
                    for with_additional_arms in [True, False]:
                        if use_model_predictions and with_additional_arms:
                            additional_arms = [
                                Arm(
                                    parameters={
                                        parameter_name: 0
                                        for parameter_name in (
                                            experiment.search_space.parameters.keys()
                                        )
                                    }
                                )
                            ]
                        else:
                            additional_arms = None

                        generation_strategy = (
                            get_default_generation_strategy_at_MBM_node(
                                experiment=experiment
                            )
                        )
                        generation_strategy.current_node._fit(experiment=experiment)
                        adapter = none_throws(generation_strategy.adapter)

                        _ = prepare_arm_data(
                            experiment=experiment,
                            metric_names=[*adapter.metric_names],
                            use_model_predictions=use_model_predictions,
                            adapter=adapter,
                            trial_index=trial_index,
                            additional_arms=additional_arms,
                        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.adapter.base import Adapter

from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCard
from ax.core.experiment import Experiment
from ax.exceptions.core import UserInputError
from ax.generation_strategy.generation_strategy import GenerationStrategy
from pyre_extensions import override


class SearchSpaceSummary(Analysis):
    """
    Creates a dataframe with information about each parameter in the given
    search space. The resulting dataframe has one row per parameter, and the
    following columns:
        - Name: the name of the parameter.
        - Type: the parameter subclass (Fixed, Range, Choice).
        - Domain: the parameter's domain (e.g., "range=[0, 1]" or
            "values=['a', 'b']").
        - Datatype: the datatype of the parameter (int, float, str, bool).
        - Flags: flags associated with the parameter, if any.
        - Target Value: the target value of the parameter, if applicable.
        - Dependent Parameters: for parameters in hierarchical search spaces,
        mapping from parameter value -> list of dependent parameter names.
    """

    @override
    def compute(
        self,
        experiment: Experiment | None = None,
        generation_strategy: GenerationStrategy | None = None,
        adapter: Adapter | None = None,
    ) -> AnalysisCard:
        if experiment is None:
            raise UserInputError(
                "`SearchSpaceSummary` analysis requires an `Experiment` input"
            )

        return self._create_analysis_card(
            title=f"SearchSpaceSummary for `{experiment.name}`",
            subtitle=(
                "The search space summary provides an overview of all "
                "parameters, including their names, types, and ranges or "
                "categories. This holistic view provides quick understanding "
                "on the parameters being optimized, allowing one to verify "
                "and adjust the search space configuration for effective "
                "exploration."
            ),
            df=experiment.search_space.summary_df,
        )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.analysis.analysis import Analysis
from ax.analysis.analysis_card import AnalysisCard
from ax.analysis.metric_summary import MetricSummary
from ax.analysis.search_space_summary import SearchSpaceSummary
from ax.analysis.summary import Summary
from ax.analysis.markdown import *  # noqa
from ax.analysis.plotly import *  # noqa

__all__ = [
    "Analysis",
    "AnalysisCard",
    "MetricSummary",
    "SearchSpaceSummary",
    "Summary",
]



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from __future__ import annotations

from collections.abc import Iterable, Mapping
from copy import deepcopy
from io import StringIO
from typing import Any, cast, TypeVar

import numpy as np
import pandas as pd
from ax.core.types import TTrialEvaluation
from ax.utils.common.base import Base
from ax.utils.common.serialization import (
    extract_init_args,
    SerializationMixin,
    serialize_init_args,
    TClassDecoderRegistry,
    TDecoderRegistry,
)
from pyre_extensions import assert_is_instance

TData = TypeVar("TData", bound="Data")
DF_REPR_MAX_LENGTH = 1000


class Data(Base, SerializationMixin):
    """Class storing numerical data for an experiment.

    The dataframe is retrieved via the `df` property. The data can be stored
    to an external store for future use by attaching it to an experiment using
    `experiment.attach_data()` (this requires a description to be set.)


    Attributes:
        df: DataFrame with underlying data, and required columns. For Data, the
            required columns are "arm_name", "metric_name", "mean", and "sem", the
            latter two of which must be numeric.

    """

    # Note: Although the SEM (standard error of the mean) is a required column,
    # downstream models can infer missing SEMs. Simply specify NaN as the SEM value,
    # either in your Metric class or in Data explicitly.
    REQUIRED_COLUMNS = {"trial_index", "arm_name", "metric_name", "mean", "sem"}

    # Note on text data: https://pandas.pydata.org/docs/user_guide/text.html
    # Its type can either be `numpy.dtypes.ObjectDType` or StringDtype extension
    # type; the later is still experimental. So we are using object.
    COLUMN_DATA_TYPES: dict[str, Any] = {
        # Ubiquitous columns.
        "trial_index": int,
        "arm_name": np.dtype("O"),
        # Metric data-related columns.
        "metric_name": np.dtype("O"),
        "mean": np.float64,
        "sem": np.float64,
        # Metadata columns available for all subclasses.
        "start_time": pd.Timestamp,
        "end_time": pd.Timestamp,
        "n": int,
        # Metadata columns available for only some subclasses.
        "frac_nonnull": np.float64,
        "random_split": int,
        "fidelities": np.dtype("O"),  # Dictionary stored as json
    }

    _df: pd.DataFrame

    def __init__(
        self: TData,
        df: pd.DataFrame | None = None,
        _skip_ordering_and_validation: bool = False,
    ) -> None:
        """Initialize a ``Data`` object from the given DataFrame.

        Args:
            df: DataFrame with underlying data, and required columns.
            _skip_ordering_and_validation: If True, uses the given DataFrame
                as is, without ordering its columns or validating its contents.
                Intended only for use in `Data.filter`, where the contents
                of the DataFrame are known to be ordered and valid.
        """
        if df is None:
            # Initialize with barebones DF.
            self._df = pd.DataFrame(columns=list(self.required_columns()))
        elif _skip_ordering_and_validation:
            self._df = df
        else:
            columns = set(df.columns)
            missing_columns = self.required_columns() - columns
            if missing_columns:
                raise ValueError(
                    f"Dataframe must contain required columns {list(missing_columns)}."
                )
            extra_columns = columns - self.supported_columns()
            if extra_columns:
                raise ValueError(f"Columns {list(extra_columns)} are not supported.")
            df = df.dropna(axis=0, how="all", ignore_index=True)
            df = self._safecast_df(df=df)

            # Reorder the columns for easier viewing
            col_order = [c for c in self.column_data_types() if c in df.columns]
            self._df = df.reindex(columns=col_order, copy=False)

    @classmethod
    def _safecast_df(
        cls: type[TData],
        df: pd.DataFrame,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        extra_column_types: Mapping[str, type] | None = None,
    ) -> pd.DataFrame:
        """Function for safely casting df to standard data types.

        Needed because numpy does not support NaNs in integer arrays.

        Allows `Any` to be specified as a type, and will skip casting for that column.

        Args:
            df: DataFrame to safe-cast.
            extra_column_types: types of columns only specified at instantiation-time.

        Returns:
            safe_df: DataFrame cast to standard dtypes.

        """
        dtypes = df.dtypes
        for col, coltype in cls.column_data_types(
            extra_column_types=extra_column_types
        ).items():
            if col in df.columns.values and coltype is not Any:
                # Pandas timestamp handlng is weird
                dtype = "datetime64[ns]" if coltype is pd.Timestamp else coltype
                if (dtype != dtypes[col]) and not (
                    coltype is int and df.loc[:, col].isnull().any()
                ):
                    df[col] = df[col].astype(dtype)
        return df

    def required_columns(self) -> set[str]:
        """Names of columns that must be present in the underlying ``DataFrame``."""
        return self.REQUIRED_COLUMNS

    @classmethod
    def supported_columns(
        cls, extra_column_names: Iterable[str] | None = None
    ) -> set[str]:
        """Names of columns supported (but not necessarily required) by this class."""
        extra_column_names = set(extra_column_names or [])
        extra_column_types: dict[str, Any] = {name: Any for name in extra_column_names}
        return cls.REQUIRED_COLUMNS.union(
            cls.column_data_types(extra_column_types=extra_column_types)
        )

    @classmethod
    def column_data_types(
        cls,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
        extra_column_types: Mapping[str, type] | None = None,
        # pyre-fixme[24]: Generic type `type` expects 1 type parameter, use
        #  `typing.Type` to avoid runtime subscripting errors.
    ) -> dict[str, type]:
        """Type specification for all supported columns."""
        extra_column_types = extra_column_types or {}
        return {**cls.COLUMN_DATA_TYPES, **extra_column_types}

    @classmethod
    def serialize_init_args(cls, obj: Any) -> dict[str, Any]:
        """Serialize the class-dependent properties needed to initialize this Data.
        Used for storage and to help construct new similar Data.
        """
        data = assert_is_instance(obj, cls)
        return serialize_init_args(
            obj=data, exclude_fields=["_skip_ordering_and_validation"]
        )

    @classmethod
    def deserialize_init_args(
        cls,
        args: dict[str, Any],
        decoder_registry: TDecoderRegistry | None = None,
        class_decoder_registry: TClassDecoderRegistry | None = None,
    ) -> dict[str, Any]:
        """Given a dictionary, extract the properties needed to initialize the object.
        Used for storage.
        """
        # Extract `df` only if present, since certain inputs to this fn, e.g.
        # SQAData.structure_metadata_json, don't have a `df` attribute.
        if "df" in args and not isinstance(args["df"], pd.DataFrame):
            # NOTE: Need dtype=False, otherwise infers arm_names like
            # "4_1" should be int 41.
            args["df"] = pd.read_json(StringIO(args["df"]["value"]), dtype=False)
        # Backward compatibility
        return extract_init_args(args=args, class_=cls)

    @property
    def true_df(self) -> pd.DataFrame:
        """Return the ``DataFrame`` being used as the source of truth (avoid using
        except for caching).
        """
        return self._df

    @property
    def df(self) -> pd.DataFrame:
        return self._df

    @staticmethod
    def from_multiple(data: Iterable[TData]) -> TData:
        """Combines multiple objects into one (with the concatenated
        underlying dataframe).

        Args:
            data: Iterable of Ax objects of this class to combine.
        """
        dfs = []

        cls = None

        for datum in data:
            if cls is None:
                cls = type(datum)
            if type(datum) is not cls:
                raise TypeError(
                    f"All data objects must be instances of the same class. Got "
                    f"{cls} and {type(datum)}."
                )
            if len(datum.df) > 0:
                dfs.append(datum.df)

        cls = cls or cast(type[TData], Data)

        if len(dfs) == 0:
            return cls()

        return cls(df=pd.concat(dfs, axis=0, sort=True))

    @classmethod
    def from_evaluations(
        cls: type[TData],
        evaluations: Mapping[str, TTrialEvaluation],
        trial_index: int,
        sample_sizes: Mapping[str, int] | None = None,
        start_time: int | str | None = None,
        end_time: int | str | None = None,
    ) -> TData:
        """
        Convert dict of evaluations to Ax data object.

        Args:
            evaluations: Map from arm name to outcomes, which itself is a mapping of
                outcome names to values, means, or tuples of mean and SEM. If SEM is
                not specified, it will be set to None and inferred from data.
            trial_index: Trial index to which this data belongs.
            sample_sizes: Number of samples collected for each arm.
            start_time: Optional start time of run of the trial that produced this
                data, in milliseconds or iso format.  Milliseconds will be automatically
                converted to iso format because iso format automatically works with the
                pandas column type `Timestamp`.
            end_time: Optional end time of run of the trial that produced this
                data, in milliseconds or iso format.  Milliseconds will be automatically
                converted to iso format because iso format automatically works with the
                pandas column type `Timestamp`.

        Returns:
            Ax object of the enclosing class.
        """
        records = cls._get_records(evaluations=evaluations, trial_index=trial_index)
        records = cls._add_cols_to_records(
            records=records,
            sample_sizes=sample_sizes,
            start_time=start_time,
            end_time=end_time,
        )
        return cls(df=pd.DataFrame(records))

    @staticmethod
    def _add_cols_to_records(
        records: list[dict[str, Any]],
        sample_sizes: Mapping[str, int] | None = None,
        start_time: int | str | None = None,
        end_time: int | str | None = None,
    ) -> list[dict[str, Any]]:
        """Adds to records metadata columns that are available for all
        Data subclasses.
        """
        if start_time is not None or end_time is not None:
            if isinstance(start_time, int):
                start_time = _ms_epoch_to_isoformat(start_time)
            if isinstance(end_time, int):
                end_time = _ms_epoch_to_isoformat(end_time)

            for record in records:
                record.update({"start_time": start_time, "end_time": end_time})
        if sample_sizes:
            for record in records:
                record["n"] = sample_sizes[str(record["arm_name"])]

        return records

    def __repr__(self) -> str:
        """String representation of the subclass, inheriting from this base."""
        df_markdown = self.df.to_markdown()
        if len(df_markdown) > DF_REPR_MAX_LENGTH:
            df_markdown = df_markdown[:DF_REPR_MAX_LENGTH] + "..."
        return f"{self.__class__.__name__}(df=\n{df_markdown})"

    @staticmethod
    def _get_records(
        evaluations: Mapping[str, TTrialEvaluation], trial_index: int
    ) -> list[dict[str, Any]]:
        return [
            {
                "arm_name": name,
                "metric_name": metric_name,
                "mean": value[0] if isinstance(value, tuple) else value,
                "sem": value[1] if isinstance(value, tuple) else None,
                "trial_index": trial_index,
            }
            for name, evaluation in evaluations.items()
            for metric_name, value in evaluation.items()
        ]

    @property
    def metric_names(self) -> set[str]:
        """Set of metric names that appear in the underlying dataframe of
        this object.
        """
        return set() if self.df.empty else set(self.df["metric_name"].values)

    def filter(
        self: TData,
        trial_indices: Iterable[int] | None = None,
        metric_names: Iterable[str] | None = None,
    ) -> TData:
        """Construct a new object with the subset of rows corresponding to the
        provided trial indices AND metric names. If either trial_indices or
        metric_names are not provided, that dimension will not be filtered.
        """

        return self.__class__(
            df=_filter_df(
                df=self.df, trial_indices=trial_indices, metric_names=metric_names
            ),
            _skip_ordering_and_validation=True,
        )

    @classmethod
    def from_multiple_data(cls, data: Iterable[Data]) -> Data:
        """Combines multiple objects into one (with the concatenated
        underlying dataframe).

        Args:
            data: Iterable of Ax objects of this class to combine.
            subset_metrics: If specified, combined object will only contain
                metrics, names of which appear in this iterable,
                in the underlying dataframe.
        """
        return cls.from_multiple(data=data)

    def clone(self) -> Data:
        """Returns a new Data object with the same underlying dataframe."""
        return Data(df=deepcopy(self.df))


def _ms_epoch_to_isoformat(epoch: int) -> str:
    return pd.Timestamp(epoch, unit="ms").isoformat()


def _filter_df(
    df: pd.DataFrame,
    trial_indices: Iterable[int] | None = None,
    metric_names: Iterable[str] | None = None,
) -> pd.DataFrame:
    """Filter rows of a dataframe by trial indices and metric names."""
    if trial_indices is not None:
        # Trial indices is not None, metric names is not yet known.
        trial_indices_mask = df["trial_index"].isin(trial_indices)
        if metric_names is None:
            # If metric names is None, we can filter & return.
            return df.loc[trial_indices_mask]
        # Both are given, filter by both.
        metric_names_mask = df["metric_name"].isin(metric_names)
        return df.loc[trial_indices_mask & metric_names_mask]
    if metric_names is not None:
        # Trial indices is None, metric names is not None.
        metric_names_mask = df["metric_name"].isin(metric_names)
        return df.loc[metric_names_mask]
    # Both are None, return the dataframe as is.
    return df



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import json
from enum import IntEnum

import pandas as pd
from ax.analysis.analysis_card import AnalysisCard


class HealthcheckStatus(IntEnum):
    PASS = 0
    FAIL = 1
    WARNING = 2


class HealthcheckAnalysisCard(AnalysisCard):
    def get_status(self) -> HealthcheckStatus:
        return HealthcheckStatus(json.loads(self.blob)["status"])

    def is_passing(self) -> bool:
        return self.get_status() == HealthcheckStatus.PASS

    def get_aditional_attrs(self) -> dict[str, str | int | float | bool]:
        return json.loads(self.blob)


def create_healthcheck_analysis_card(
    name: str,
    title: str,
    subtitle: str,
    df: pd.DataFrame,
    status: HealthcheckStatus,
    **additional_attrs: str | int | float | bool,
) -> HealthcheckAnalysisCard:
    return HealthcheckAnalysisCard(
        name=name,
        title=title,
        subtitle=subtitle,
        df=df,
        blob=json.dumps(
            {
                "status": status,
                **additional_attrs,
            }
        ),
    )



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import os

from ax.benchmark.benchmark_metric import BenchmarkMetric
from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.problems.surrogate.hss.base import HierarchicalSearchSpaceSurrogate
from ax.core.objective import Objective
from ax.core.optimization_config import OptimizationConfig
from ax.core.parameter import (
    ChoiceParameter,
    FixedParameter,
    ParameterType,
    RangeParameter,
)

from ax.core.search_space import HierarchicalSearchSpace
from xgboost import XGBRegressor


CIFAR10_SURROGATE_BASELINE_VALUE = 0.7594003677368164
CIFAR10_SURROGATE_OPTIMAL_VALUE = 0.82043886


def get_cifar10_surrogate_search_space() -> HierarchicalSearchSpace:
    search_space = HierarchicalSearchSpace(
        parameters=[
            FixedParameter(
                name="root",
                value=True,
                parameter_type=ParameterType.BOOL,
                dependents={
                    True: [
                        "lr",
                        "two-more-blocks",
                        "use_softplus_activation",
                        "use_weight_decay",
                    ],
                },
            ),
            RangeParameter(
                name="lr",
                parameter_type=ParameterType.FLOAT,
                lower=1e-5,
                upper=1e-1,
                log_scale=True,
            ),
            ChoiceParameter(
                name="two-more-blocks",
                parameter_type=ParameterType.BOOL,
                values=[False, True],
                is_ordered=True,
                sort_values=True,
            ),
            ChoiceParameter(
                name="use_softplus_activation",
                parameter_type=ParameterType.BOOL,
                values=[False, True],
                is_ordered=True,
                sort_values=True,
                dependents={
                    False: [],
                    True: ["softplus_beta"],
                },
            ),
            RangeParameter(
                name="softplus_beta",
                parameter_type=ParameterType.FLOAT,
                lower=0.5,
                upper=2.0,
            ),
            ChoiceParameter(
                "use_weight_decay",
                parameter_type=ParameterType.BOOL,
                values=[False, True],
                dependents={
                    False: [],
                    True: ["weight_decay"],
                },
                is_ordered=True,
                sort_values=True,
            ),
            RangeParameter(
                name="weight_decay",
                parameter_type=ParameterType.FLOAT,
                lower=1e-8,
                upper=1e-2,
                log_scale=True,
            ),
        ],
    )

    return search_space


def get_cifar10_surrogate_arguments() -> (
    tuple[list[list[str]], list[str], list[dict[str, bool]], list[XGBRegressor]]
):
    """
    Construct the arguments to be passed to `HierarchicalSearchSpaceSurrogate` that
    creates a surrogate model on CIFAR10.

    Examples:
    >>> lst_active_param_names, flag_param_names, lst_flag_config, lst_xgb_models = (
    ...     get_fashion_mnist_surrogate_arguments()
    ... )
    >>> surrogate = HierarchicalSearchSpaceSurrogate(
    ...     outcome_names=["test accuracy"],
    ...     lst_active_param_names=lst_active_param_names,
    ...     flag_param_names=flag_param_names,
    ...     lst_flag_config=lst_flag_config,
    ...     lst_xgb_models=lst_xgb_models,
    ... )
    """
    lst_active_param_names: list[list[str]] = [
        ["lr", "two-more-blocks"],
        ["lr", "two-more-blocks", "weight_decay"],
        ["lr", "two-more-blocks", "softplus_beta"],
        ["lr", "two-more-blocks", "softplus_beta", "weight_decay"],
    ]

    flag_param_names: list[str] = [
        "use_softplus_activation",
        "use_weight_decay",
    ]

    lst_flag_config: list[dict[str, bool]] = [
        {"use_softplus_activation": False, "use_weight_decay": False},
        {"use_softplus_activation": False, "use_weight_decay": True},
        {"use_softplus_activation": True, "use_weight_decay": False},
        {"use_softplus_activation": True, "use_weight_decay": True},
    ]

    lst_xgb_models: list[XGBRegressor] = [XGBRegressor() for i in range(4)]

    for i, xgb_model in enumerate(lst_xgb_models):
        xgb_model.load_model(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "checkpoints",
                f"cifar10_xgb_model_{i:d}.json",
            )
        )

    return lst_active_param_names, flag_param_names, lst_flag_config, lst_xgb_models


def get_cifar10_surrogate_benchmark(
    num_trials: int = 50,
) -> BenchmarkProblem:
    """Create a BenchmarkProblem for tuning hyperparameters of a residual network on
    CIFAR10. The tunable hyperparameters are as follows:
    1. The learning rate;
    2. Whether to use two additional residual blocks;
    3. Whether to use the softplus activation (instead of ReLU), and the coefficient
    beta in the softplus activation;
    4. Whether to use weight decay, and the weight decay coefficient.
    More details on the tunable hyperparameters can be found in the return value of
    `get_cifar10_surrogate_search_space`. The returned benchmark problem uses
    gradient-boosted trees to predict the validation accuracy, which makes the oracle
    computationally cheap to evaluate.
    """
    search_space = get_cifar10_surrogate_search_space()

    lst_active_param_names, flag_param_names, lst_flag_config, lst_xgb_models = (
        get_cifar10_surrogate_arguments()
    )

    test_function = HierarchicalSearchSpaceSurrogate(
        outcome_names=["CIFAR10 Test Accuracy"],
        lst_active_param_names=lst_active_param_names,
        flag_param_names=flag_param_names,
        lst_flag_config=lst_flag_config,
        lst_xgb_models=lst_xgb_models,
    )

    optimization_config = OptimizationConfig(
        objective=Objective(
            metric=BenchmarkMetric(
                name="CIFAR10 Test Accuracy",
                lower_is_better=False,
                observe_noise_sd=False,
            ),
            minimize=False,
        )
    )

    return BenchmarkProblem(
        name="CIFAR10 Test Accuracy",
        search_space=search_space,
        optimization_config=optimization_config,
        num_trials=num_trials,
        optimal_value=CIFAR10_SURROGATE_OPTIMAL_VALUE,
        test_function=test_function,
        baseline_value=CIFAR10_SURROGATE_BASELINE_VALUE,
    )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


import unittest

from ax.adapter.transforms.base import Transform
from ax.adapter.transforms.choice_encode import (
    ChoiceEncode,
    ChoiceToNumericChoice,
    OrderedChoiceEncode,
    OrderedChoiceToIntegerRange,
)
from ax.adapter.transforms.int_range_to_choice import IntRangeToChoice
from ax.adapter.transforms.map_key_to_float import MapKeyToFloat
from ax.adapter.transforms.task_encode import TaskChoiceToIntTaskChoice, TaskEncode
from ax.storage.json_store.decoder import object_from_json
from ax.storage.json_store.decoders import transform_type_from_json
from ax.storage.json_store.encoder import object_to_json
from ax.storage.json_store.encoders import transform_type_to_dict
from ax.storage.transform_registry import register_transform


class TestTransformEncode(unittest.TestCase):
    def setUp(self) -> None:
        super().setUp()

        self.deprecatedTestCases = [
            (OrderedChoiceEncode, OrderedChoiceToIntegerRange),
            (ChoiceEncode, ChoiceToNumericChoice),
            (TaskEncode, TaskChoiceToIntTaskChoice),
        ]

    def test_encode_and_decode_transform(self) -> None:
        with self.subTest("IntRangeToChoice"):
            transform_dict = transform_type_to_dict(IntRangeToChoice)
            self.assertEqual(transform_dict["__type"], "Type[Transform]")
            self.assertIn("IntRangeToChoice", transform_dict["transform_type"])

            decoded_transform_type = transform_type_from_json(transform_dict)
            self.assertEqual(decoded_transform_type, IntRangeToChoice)

        with self.subTest("MapKeyToFloat"):
            transform_dict = transform_type_to_dict(MapKeyToFloat)
            self.assertEqual(transform_dict["__type"], "Type[Transform]")
            self.assertIn("MapKeyToFloat", transform_dict["transform_type"])

            decoded_transform_type = transform_type_from_json(transform_dict)
            self.assertEqual(decoded_transform_type, MapKeyToFloat)

    def test_encode_and_decode_deprecated_transforms(self) -> None:
        for deprecated_type, current_type in self.deprecatedTestCases:
            transform_dict = transform_type_to_dict(deprecated_type)
            self.assertEqual(transform_dict["__type"], "Type[Transform]")
            self.assertIn(deprecated_type.__name__, transform_dict["transform_type"])

            decoded_transform_type = transform_type_from_json(transform_dict)
            # Deprecated type is decoded into the current type.
            self.assertNotEqual(decoded_transform_type, deprecated_type)
            self.assertEqual(decoded_transform_type, current_type)

    def test_register_transform(self) -> None:
        class DummyTransform(Transform):
            pass

        with self.assertRaisesRegex(KeyError, "DummyTransform"):
            object_from_json(object_to_json(DummyTransform))

        register_transform(DummyTransform)
        self.assertIs(object_from_json(object_to_json(DummyTransform)), DummyTransform)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from ax.benchmark.benchmark_metric import BenchmarkMetric
from ax.benchmark.benchmark_test_functions.surrogate import SurrogateTestFunction
from ax.benchmark.problems.surrogate.lcbench.transfer_learning import (
    DEFAULT_AND_OPTIMAL_VALUES,
    get_lcbench_benchmark_problem,
)
from ax.core.observation import ObservationFeatures
from ax.utils.common.testutils import TestCase
from pyre_extensions import assert_is_instance


class TestLCBenchBenchmark(TestCase):
    @TestCase.ax_long_test(reason="Training random forest regressor")
    def test_lcbench_predictions(self) -> None:
        self.assertEqual(len(DEFAULT_AND_OPTIMAL_VALUES), 22)
        # NOTE: lots of tasks, so testing only one here o/w this is very slow
        dataset_name = "car"
        problem = get_lcbench_benchmark_problem(
            dataset_name=dataset_name,
            num_trials=32,
        )
        test_function = assert_is_instance(problem.test_function, SurrogateTestFunction)
        metric = assert_is_instance(
            problem.optimization_config.objective.metric, BenchmarkMetric
        )
        self.assertFalse(metric.observe_noise_sd)
        self.assertEqual(problem.num_trials, 32)
        default_val, opt_val = DEFAULT_AND_OPTIMAL_VALUES[dataset_name]
        self.assertAlmostEqual(
            float(problem.optimal_value),
            opt_val,
            places=4,
        )
        surrogate = test_function.surrogate

        # Predict for arm 0_0 and make sure it matches the expected value
        training_data = surrogate.get_training_data()
        features = training_data.arm_data.loc[0, "0_0"].to_dict()
        pred, _ = surrogate.predict(
            observation_features=[ObservationFeatures(parameters=features)]
        )
        self.assertAlmostEqual(pred["Train/val_accuracy"][0], default_val, places=3)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

import functools
import warnings
from collections.abc import Callable
from copy import deepcopy
from functools import partial
from typing import Any

import torch
from ax.core.search_space import SearchSpaceDigest
from ax.exceptions.core import AxWarning
from ax.generators.torch.botorch_modular.acquisition import Acquisition
from ax.generators.torch.botorch_modular.optimizer_argparse import optimizer_argparse
from ax.generators.torch.botorch_modular.surrogate import Surrogate
from ax.generators.torch_base import TorchOptConfig
from botorch.acquisition.acquisition import AcquisitionFunction
from botorch.acquisition.multi_objective.logei import (
    qLogNoisyExpectedHypervolumeImprovement,
)
from botorch.acquisition.penalized import L0Approximation
from botorch.models.deterministic import GenericDeterministicModel
from botorch.models.model import ModelList
from botorch.optim import (
    gen_batch_initial_conditions,
    Homotopy,
    HomotopyParameter,
    LogLinearHomotopySchedule,
    optimize_acqf_homotopy,
)
from botorch.utils.datasets import SupervisedDataset
from pyre_extensions import assert_is_instance, none_throws
from torch import Tensor

CLAMP_TOL = 1e-2


class SEBOAcquisition(Acquisition):
    """
    Implement the acquisition function of Sparsity Exploring Bayesian
    Optimization (SEBO).

    The SEBO is a hyperparameter-free method to simultaneously maximize a target
    objective and sparsity. When L0 norm is used, SEBO uses a novel differentiable
    relaxation based on homotopy continuation to efficiently optimize for sparsity.
    """

    def __init__(
        self,
        surrogate: Surrogate,
        search_space_digest: SearchSpaceDigest,
        torch_opt_config: TorchOptConfig,
        botorch_acqf_class: type[AcquisitionFunction],
        botorch_acqf_options: dict[str, Any],
        botorch_acqf_classes_with_options: list[
            tuple[type[AcquisitionFunction], dict[str, Any]]
        ]
        | None = None,
        n: int | None = None,
        options: dict[str, Any] | None = None,
    ) -> None:
        tkwargs: dict[str, Any] = {"dtype": surrogate.dtype, "device": surrogate.device}
        options = options or {}
        botorch_acqf_options = botorch_acqf_options or {}
        self.penalty_name: str = options.get("penalty", "L0_norm")
        self.target_point: Tensor = options.get("target_point", None)
        if self.target_point is None:
            raise ValueError("please provide target point.")
        self.target_point.to(**tkwargs)
        self.sparsity_threshold: int = options.get(
            "sparsity_threshold", surrogate.Xs[0].shape[-1]
        )
        # construct deterministic model for penalty term
        # pyre-fixme[4]: Attribute must be annotated.
        self.deterministic_model = self._construct_penalty()
        surrogate_f = deepcopy(surrogate)
        # we need to clamp the training data to the target point here as it may
        # be slightly off due to numerical issues.
        X_sparse = clamp_to_target(
            X=surrogate_f.Xs[0].clone(),
            target_point=self.target_point,
            clamp_tol=CLAMP_TOL,
        )
        # update the training data in new surrogate
        none_throws(surrogate_f._training_data).append(
            SupervisedDataset(
                X=X_sparse,
                Y=self.deterministic_model(X_sparse),
                Yvar=torch.zeros(X_sparse.shape[0], 1, **tkwargs),  # noiseless
                feature_names=surrogate_f.training_data[0].feature_names,
                outcome_names=[self.penalty_name],
            )
        )
        # update the model in new surrogate
        surrogate_f._model = ModelList(surrogate.model, self.deterministic_model)
        # update objective weights and thresholds in the torch config
        torch_opt_config_sebo = self._transform_torch_config(
            torch_opt_config=torch_opt_config, **tkwargs
        )

        # Change some options (note: we do not want to do this in-place)
        if botorch_acqf_options.get("cache_root", False):
            warnings.warn(
                "SEBO doesn't support `cache_root=True`. Changing it to `False`.",
                AxWarning,
                stacklevel=3,
            )
            botorch_acqf_options = {**botorch_acqf_options, "cache_root": False}

        # Instantiate the `botorch_acqf_class`. We need to modify `a` before doing this
        # (as it controls the L0 norm approximation) since the baseline will be pruned
        # when the acquisition function is created. With a=1e-6 the deterministic model
        # will be numerically close to the true L0 norm and we will select the
        # baseline according to the last homotopy step.
        if self.penalty_name == "L0_norm":
            self.deterministic_model._f.a.fill_(1e-6)
        super().__init__(
            surrogate=surrogate_f,
            search_space_digest=search_space_digest,
            torch_opt_config=torch_opt_config_sebo,
            botorch_acqf_class=qLogNoisyExpectedHypervolumeImprovement,
            botorch_acqf_options=botorch_acqf_options,
            options=options,
        )

        # update objective threshold for deterministic model (penalty term)
        # pyre-fixme[29]: `Union[(self: TensorBase, indices: Union[None, slice[Any, A...
        self.acqf.ref_point[-1] = self.sparsity_threshold * -1
        self._objective_thresholds[-1] = self.sparsity_threshold  # pyre-ignore

    def _construct_penalty(self) -> GenericDeterministicModel:
        """Construct a penalty term as deterministic model to be included in
        SEBO acqusition function. Currently only L0 and L1 penalty are supported.
        """
        if self.penalty_name == "L0_norm":
            L0 = L0Approximation(target_point=self.target_point)
            return GenericDeterministicModel(f=L0)
        elif self.penalty_name == "L1_norm":
            L1 = functools.partial(
                L1_norm_func,
                init_point=self.target_point,
            )
            return GenericDeterministicModel(f=L1)
        else:
            raise NotImplementedError(
                f"{self.penalty_name} is not currently implemented."
            )

    def _transform_torch_config(
        self,
        torch_opt_config: TorchOptConfig,
        **tkwargs: Any,
    ) -> TorchOptConfig:
        """Transform torch config to include penalty term (deterministic model) as
        an additional outcomes in BoTorch model.
        """
        # update objective weights by appending the weight -1 for sparsity objective.
        objective_weights_sebo = torch.cat(
            [torch_opt_config.objective_weights, -torch.ones(1, **tkwargs)]
        )
        if torch_opt_config.outcome_constraints is not None:
            # update the shape of A matrix in outcome_constraints
            A, b = none_throws(torch_opt_config.outcome_constraints)
            outcome_constraints_sebo = (
                torch.cat([A, torch.zeros(A.shape[0], 1, **tkwargs)], dim=1),
                b,
            )
        else:
            outcome_constraints_sebo = None
        if torch_opt_config.objective_thresholds is not None:
            objective_thresholds_sebo = torch.cat(
                [
                    torch_opt_config.objective_thresholds,
                    torch.tensor([self.sparsity_threshold], **tkwargs),
                ]
            )
        else:
            # NOTE: The reference point will be inferred in the base class.
            objective_thresholds_sebo = None

        # update pending observations (if not none) by appending an obs for
        # the new penalty outcome
        pending_observations = torch_opt_config.pending_observations
        if torch_opt_config.pending_observations is not None:
            pending_observations = torch_opt_config.pending_observations + [
                torch_opt_config.pending_observations[0]
            ]

        return TorchOptConfig(
            objective_weights=objective_weights_sebo,
            outcome_constraints=outcome_constraints_sebo,
            objective_thresholds=objective_thresholds_sebo,
            linear_constraints=torch_opt_config.linear_constraints,
            fixed_features=torch_opt_config.fixed_features,
            pending_observations=pending_observations,
            model_gen_options=torch_opt_config.model_gen_options,
            rounding_func=torch_opt_config.rounding_func,
            opt_config_metrics=torch_opt_config.opt_config_metrics,
            is_moo=True,  # SEBO adds an objective, so it'll always be MOO.
        )

    def optimize(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
        fixed_features: dict[int, float] | None = None,
        rounding_func: Callable[[Tensor], Tensor] | None = None,
        optimizer_options: dict[str, Any] | None = None,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Generate a set of candidates via multi-start optimization. Obtains
        candidates and their associated acquisition function values.

        Args:
            n: The number of candidates to generate.
            search_space_digest: A ``SearchSpaceDigest`` object containing search space
                properties, e.g. ``bounds`` for optimization.
            inequality_constraints: A list of tuples (indices, coefficients, rhs),
                with each tuple encoding an inequality constraint of the form
                ``sum_i (X[indices[i]] * coefficients[i]) >= rhs``.
            fixed_features: A map `{feature_index: value}` for features that
                should be fixed to a particular value during generation.
            rounding_func: A function that post-processes an optimization
                result appropriately (i.e., according to `round-trip`
                transformations).
            optimizer_options: Options for the optimizer function, e.g. ``sequential``
                or ``raw_samples``.

        Returns:
            A three-element tuple containing an `n x d`-dim tensor of generated
            candidates, a tensor with the associated acquisition values, and a tensor
            with the weight for each candidate.
        """
        if self.penalty_name == "L0_norm":
            candidates, expected_acquisition_value, weights = (
                self._optimize_with_homotopy(
                    n=n,
                    search_space_digest=search_space_digest,
                    inequality_constraints=inequality_constraints,
                    fixed_features=fixed_features,
                    rounding_func=rounding_func,
                    optimizer_options=optimizer_options,
                )
            )
        else:
            # if L1 norm use standard moo-opt
            candidates, expected_acquisition_value, weights = super().optimize(
                n=n,
                search_space_digest=search_space_digest,
                inequality_constraints=inequality_constraints,
                fixed_features=fixed_features,
                rounding_func=rounding_func,
                optimizer_options=optimizer_options,
            )

        # similar, make sure if applies to sparse dimensions only
        candidates = clamp_to_target(
            X=candidates, target_point=self.target_point, clamp_tol=CLAMP_TOL
        )
        return candidates, expected_acquisition_value, weights

    def _optimize_with_homotopy(
        self,
        n: int,
        search_space_digest: SearchSpaceDigest,
        inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
        fixed_features: dict[int, float] | None = None,
        rounding_func: Callable[[Tensor], Tensor] | None = None,
        optimizer_options: dict[str, Any] | None = None,
    ) -> tuple[Tensor, Tensor, Tensor]:
        """Optimize SEBO ACQF with L0 norm using homotopy."""
        optimizer_options = optimizer_options or {}
        # extend to fixed a no homotopy_schedule schedule
        _tensorize = partial(torch.tensor, dtype=self.dtype, device=self.device)
        bounds = _tensorize(search_space_digest.bounds).t()
        homotopy_schedule = LogLinearHomotopySchedule(
            start=0.2,
            end=1e-3,
            num_steps=optimizer_options.get("num_homotopy_steps", 30),
        )

        # Prepare arguments for optimizer
        optimizer_options_with_defaults = optimizer_argparse(
            self.acqf,
            optimizer_options=optimizer_options,
            optimizer="optimize_acqf_homotopy",
        )

        homotopy = Homotopy(
            homotopy_parameters=[
                HomotopyParameter(
                    parameter=self.deterministic_model._f.a,
                    schedule=homotopy_schedule,
                )
            ],
        )

        if "batch_initial_conditions" not in optimizer_options_with_defaults:
            optimizer_options_with_defaults["batch_initial_conditions"] = (
                get_batch_initial_conditions(
                    acq_function=self.acqf,
                    raw_samples=optimizer_options_with_defaults["raw_samples"],
                    inequality_constraints=inequality_constraints,
                    fixed_features=fixed_features,
                    X_pareto=assert_is_instance(self.acqf.X_baseline, Tensor),
                    target_point=self.target_point,
                    bounds=bounds,
                    num_restarts=optimizer_options_with_defaults["num_restarts"],
                )
            )

        candidates, expected_acquisition_value = optimize_acqf_homotopy(
            q=n,
            acq_function=self.acqf,
            bounds=bounds,
            homotopy=homotopy,
            num_restarts=optimizer_options_with_defaults["num_restarts"],
            raw_samples=optimizer_options_with_defaults["raw_samples"],
            inequality_constraints=inequality_constraints,
            post_processing_func=rounding_func,
            fixed_features=fixed_features,
            batch_initial_conditions=optimizer_options_with_defaults[
                "batch_initial_conditions"
            ],
        )
        return (
            candidates,
            expected_acquisition_value,
            torch.ones(n, device=candidates.device, dtype=candidates.dtype),
        )


def L1_norm_func(X: Tensor, init_point: Tensor) -> Tensor:
    r"""L1_norm takes in a a `batch_shape x n x d`-dim input tensor `X`
    to a `batch_shape x n x 1`-dimensional L1 norm tensor. To be used
    for constructing a GenericDeterministicModel.
    """
    return torch.linalg.norm((X - init_point), ord=1, dim=-1, keepdim=True)


def clamp_to_target(X: Tensor, target_point: Tensor, clamp_tol: float) -> Tensor:
    """Clamp generated candidates within the given ranges to the target point.

    Args:
        X: A `batch_shape x n x d`-dim input tensor `X`.
        target_point: A tensor of size `d` corresponding to the target point.
        clamp_tol: The clamping tolerance. Any value within `clamp_tol` of the
            `target_point` will be clamped to the `target_point`.
    """
    clamp_mask = (X - target_point).abs() <= clamp_tol
    X[clamp_mask] = target_point.clone().repeat(*X.shape[:-1], 1)[clamp_mask]
    return X


def get_batch_initial_conditions(
    acq_function: AcquisitionFunction,
    raw_samples: int,
    X_pareto: Tensor,
    target_point: Tensor,
    bounds: Tensor,
    num_restarts: int = 20,
    inequality_constraints: list[tuple[Tensor, Tensor, float]] | None = None,
    fixed_features: dict[int, float] | None = None,
) -> Tensor:
    """Generate starting points for the SEBO acquisition function optimization."""
    tkwargs: dict[str, Any] = {"device": X_pareto.device, "dtype": X_pareto.dtype}
    num_rand = num_restarts if len(X_pareto) == 0 else num_restarts // 2
    num_local = num_restarts - num_rand

    # (1) Random points (Sobol if no constraints, otherwise uses hit-and-run)
    X_cand_rand = gen_batch_initial_conditions(
        acq_function=acq_function,
        bounds=bounds,
        q=1,
        raw_samples=raw_samples,
        num_restarts=num_rand,
        options={"topn": True},
        fixed_features=fixed_features,
        inequality_constraints=inequality_constraints,
    ).to(**tkwargs)

    if num_local == 0:
        return X_cand_rand

    # (2) Perturbations of points on the Pareto frontier (done by TuRBO/Spearmint)
    X_cand_local = X_pareto.clone()[
        torch.randint(high=len(X_pareto), size=(raw_samples,))
    ]
    mask = X_cand_local != target_point
    X_cand_local[mask] += (
        0.2 * ((bounds[1] - bounds[0]) * torch.randn_like(X_cand_local))[mask]
    )
    X_cand_local = torch.clamp(X_cand_local.unsqueeze(1), min=bounds[0], max=bounds[1])
    X_cand_local = X_cand_local[acq_function(X_cand_local).topk(num_local).indices]
    return torch.cat((X_cand_rand, X_cand_local), dim=0)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from copy import deepcopy
from unittest import mock
from unittest.mock import MagicMock

from ax.adapter.random import RandomAdapter

from ax.core.arm import Arm
from ax.core.data import Data
from ax.core.experiment import Experiment
from ax.core.generator_run import GeneratorRun
from ax.core.observation import ObservationFeatures
from ax.core.types import TParameterization
from ax.exceptions.core import UnsupportedError
from ax.generation_strategy.external_generation_node import ExternalGenerationNode
from ax.generation_strategy.generation_strategy import GenerationStrategy
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import (
    get_branin_data,
    get_branin_experiment,
    get_sobol,
)
from pyre_extensions import none_throws


class DummyNode(ExternalGenerationNode):
    def __init__(self, should_deduplicate: bool = False) -> None:
        super().__init__(node_name="dummy", should_deduplicate=should_deduplicate)
        self.update_count = 0
        self.gen_count = 0
        self.generator: RandomAdapter | None = None
        self.last_pending: list[TParameterization] = []

    def update_generator_state(self, experiment: Experiment, data: Data) -> None:
        self.update_count += 1
        self.generator = get_sobol(experiment.search_space)

    def get_next_candidate(
        self, pending_parameters: list[TParameterization]
    ) -> TParameterization:
        self.gen_count += 1
        self.last_pending = deepcopy(pending_parameters)
        return none_throws(self.generator).gen(n=1).arms[0].parameters


class TestExternalGenerationNode(TestCase):
    def test_properties(self) -> None:
        node = DummyNode()
        self.assertEqual(node.node_name, "dummy")
        self.assertGreater(node.fit_time_since_gen, 0.0)
        self.assertIsNone(node._fitted_adapter)
        self.assertIsNone(node.generator_spec_to_gen_from)
        self.assertFalse(node.should_deduplicate)
        with self.assertRaisesRegex(UnsupportedError, "Unexpected arguments"):
            node._fit(
                experiment=MagicMock(), data=MagicMock(), search_space=MagicMock()
            )

    def test_generation(self) -> None:
        node = DummyNode()
        gs = GenerationStrategy(name="test gs", nodes=[node])
        self.assertEqual(node.gen_count, 0)
        self.assertEqual(node.update_count, 0)
        experiment = get_branin_experiment()

        # Sequential generation.
        for _ in range(3):
            gr = gs.gen_single_trial(
                n=1, experiment=experiment, data=experiment.lookup_data()
            )
            trial = experiment.new_trial(generator_run=gr)
            trial.mark_running(no_runner_required=True)
            experiment.attach_data(get_branin_data(trials=[trial]))
            trial.mark_completed()
        self.assertEqual(node.gen_count, 3)
        self.assertEqual(node.update_count, 3)
        self.assertEqual(node.last_pending, [])

        # Test pending point handling.
        pending_observations = {
            "some_metric": [ObservationFeatures(parameters={"x1": 0.123, "x2": 0.456})]
        }
        gr = gs.gen_single_trial(
            n=1,
            experiment=experiment,
            data=experiment.lookup_data(),
            pending_observations=pending_observations,
        )
        trial = experiment.new_trial(generator_run=gr)
        trial.mark_running(no_runner_required=True)
        experiment.attach_data(get_branin_data(trials=[trial]))
        trial.mark_completed()
        self.assertEqual(node.gen_count, 4)
        self.assertEqual(node.update_count, 4)
        self.assertEqual(node.last_pending, [{"x1": 0.123, "x2": 0.456}])

        # Batch generation.
        gr = gs.gen_single_trial(
            n=5, experiment=experiment, data=experiment.lookup_data()
        )
        self.assertEqual(node.gen_count, 9)
        self.assertEqual(node.update_count, 5)
        self.assertEqual(len(gr.arms), 5)
        self.assertGreater(none_throws(gr.fit_time), 0.0)
        self.assertGreater(none_throws(gr.gen_time), 0.0)
        self.assertEqual(gr._model_key, "dummy")
        self.assertEqual(len(node.last_pending), 4)

    def test_fallback(self) -> None:
        node = DummyNode(should_deduplicate=True)
        gs = GenerationStrategy(name="test gs", nodes=[node])
        experiment = get_branin_experiment()
        params = {"x1": 0.0, "x2": 0.0}
        with mock.patch(
            "ax.adapter.random.RandomAdapter.gen",
            return_value=GeneratorRun(arms=[Arm(parameters=params)], model_key="Sobol"),
        ) as mock_gen:
            experiment.new_trial(
                generator_run=gs.gen_single_trial(n=1, experiment=experiment)
            ).run()
            experiment.new_trial(
                generator_run=gs.gen_single_trial(n=1, experiment=experiment)
            ).run()
        self.assertEqual(node.gen_count, 6)  # first trial + 5 deduplication attempts
        self.assertEqual(mock_gen.call_count, 7)  # also called for the fallback
        self.assertEqual(node.update_count, 2)
        # First trial was generated by the node, second by the fallback.
        self.assertEqual(experiment.trials[0].generator_runs[0]._model_key, "dummy")
        self.assertEqual(experiment.trials[1].generator_runs[0]._model_key, "Sobol")



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from datetime import datetime, timedelta

from ax.analysis.healthcheck.can_generate_candidates import (
    CanGenerateCandidatesAnalysis,
)
from ax.analysis.healthcheck.healthcheck_analysis import HealthcheckStatus
from ax.core.trial_status import TrialStatus
from ax.utils.common.testutils import TestCase
from ax.utils.testing.core_stubs import get_branin_experiment


class TestCanGenerateCandidates(TestCase):
    def test_passes_if_can_generate(self) -> None:
        # GIVEN we can generate candidates
        # WHEN we run the healthcheck
        card = CanGenerateCandidatesAnalysis(
            can_generate_candidates=True,
            reason="No problems found.",
            days_till_fail=0,
        ).compute(experiment=None, generation_strategy=None)
        # THEN it is PASSES
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)
        self.assertEqual(card.name, "CanGenerateCandidatesAnalysis")
        self.assertEqual(card.title, "Ax Candidate Generation Success")
        self.assertEqual(
            card.subtitle,
            (
                "The candidate generation health check notifies users if key "
                "criteria for candidate generation are missing. No problems found."
            ),
        )
        self.assertEqual(card.get_status(), HealthcheckStatus.PASS)
        self.assertDictEqual(
            card.get_aditional_attrs(),
            {
                "status": HealthcheckStatus.PASS,
                "reason": "No problems found.",
            },
        )

    def test_warns_if_a_trial_was_recently_run(self) -> None:
        # GIVEN an experiment with a recently run trial
        experiment = get_branin_experiment(with_trial=True)
        trial = experiment.trials[0]
        trial.mark_running(no_runner_required=True)
        trial._time_run_started = datetime.now() - timedelta(days=1)
        # WHEN we run the healthcheck
        card = CanGenerateCandidatesAnalysis(
            can_generate_candidates=False,
            reason="The data is borked.",
            days_till_fail=2,
        ).compute(experiment=experiment, generation_strategy=None)
        # THEN it is a WARNING
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)
        self.assertEqual(card.name, "CanGenerateCandidatesAnalysis")
        self.assertEqual(card.title, "Ax Candidate Generation Warning")
        self.assertEqual(
            card.subtitle,
            (
                "The candidate generation health check notifies users if key "
                "criteria for candidate generation are missing. "
                f"{CanGenerateCandidatesAnalysis.REASON_PREFIX}"
                "The data is borked.\n\n"
                "LAST TRIAL RUN: 1 day(s) ago"
            ),
        )
        self.assertEqual(card.get_status(), HealthcheckStatus.WARNING)
        self.assertDictEqual(
            card.get_aditional_attrs(),
            {
                "status": HealthcheckStatus.WARNING,
                "reason": "The data is borked.",
            },
        )

    def test_is_fail_no_trials_have_been_run(self) -> None:
        # GIVEN an experiment with a candidate trial
        experiment = get_branin_experiment(with_trial=True)
        trial = experiment.trials[0]
        self.assertEqual(trial.status, TrialStatus.CANDIDATE)
        # WHEN we run the healthcheck
        card = CanGenerateCandidatesAnalysis(
            can_generate_candidates=False,
            reason="The data is gone.",
            days_till_fail=2,
        ).compute(experiment=experiment, generation_strategy=None)
        # THEN it is an ERROR
        self.assertEqual(card.get_status(), HealthcheckStatus.FAIL)
        self.assertEqual(card.name, "CanGenerateCandidatesAnalysis")
        self.assertEqual(card.title, "Ax Candidate Generation Failure")
        self.assertEqual(
            card.subtitle,
            (
                "The candidate generation health check notifies users if key "
                "criteria for candidate generation are missing. "
                f"{CanGenerateCandidatesAnalysis.REASON_PREFIX}The data is gone."
            ),
        )
        self.assertEqual(card.get_status(), HealthcheckStatus.FAIL)
        self.assertDictEqual(
            card.get_aditional_attrs(),
            {
                "status": HealthcheckStatus.FAIL,
                "reason": "The data is gone.",
            },
        )

    def test_is_fail_if_no_trial_was_recently_run(self) -> None:
        # GIVEN an experiment with an old trial
        experiment = get_branin_experiment(with_trial=True)
        trial = experiment.trials[0]
        trial.mark_running(no_runner_required=True)
        trial._time_run_started = datetime.now() - timedelta(days=3)
        trial.mark_completed()
        # WHEN we run the healthcheck
        card = CanGenerateCandidatesAnalysis(
            can_generate_candidates=False,
            reason="The data is old.",
            days_till_fail=1,
        ).compute(experiment=experiment, generation_strategy=None)
        # THEN it is an ERROR
        self.assertEqual(card.get_status(), HealthcheckStatus.FAIL)
        self.assertEqual(card.name, "CanGenerateCandidatesAnalysis")
        self.assertEqual(card.title, "Ax Candidate Generation Failure")
        self.assertEqual(
            card.subtitle,
            (
                "The candidate generation health check notifies users if key "
                "criteria for candidate generation are missing. "
                f"{CanGenerateCandidatesAnalysis.REASON_PREFIX}"
                "The data is old.\n\n"
                "LAST TRIAL RUN: 3 day(s) ago"
            ),
        )
        self.assertEqual(card.get_status(), HealthcheckStatus.FAIL)
        self.assertDictEqual(
            card.get_aditional_attrs(),
            {
                "status": HealthcheckStatus.FAIL,
                "reason": "The data is old.",
            },
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict

from warnings import catch_warnings

import numpy as np
from ax.generators.discrete.eb_ashr import EBAshr, no_model_feasibility_util
from ax.utils.common.testutils import TestCase


class EBAshrTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.Xs = [
            [[1, 1], [2, 2], [3, 3], [4, 4], [5, 5]]
        ]  # 5 arms, each of dimensionality 2
        self.Ys = [
            [1, -1, 2, -2, 0],
            [-1, 1, -2, 2, 0],
        ]  # two metrics evaluated on each of the five arms
        self.Yvars = [[1, 1, 1, 1, 1], [1, 1, 1, 1, 1]]
        self.parameter_values = [
            [1, 2, 3, 4, 5],
            [1, 2, 3, 4, 5],
        ]  # not used in best arm selection
        self.outcome_names = ["x", "y"]  # metric names, not used in best arm selection

    def test_EBAshr(self) -> None:
        generator = EBAshr()
        self.assertFalse(generator.can_predict)
        self.assertTrue(generator.can_model_in_sample)
        generator.fit(
            Xs=self.Xs,
            Ys=self.Ys,
            Yvars=self.Yvars,
            parameter_values=self.parameter_values,
            outcome_names=self.outcome_names,
        )
        arms, weights, gen_metadata = generator.gen(
            n=3,
            parameter_values=self.parameter_values,
            outcome_constraints=(np.array([[0.0, -1.0]]), np.array([[0.0]])),
            objective_weights=np.array([1, 0]),
        )

        self.assertEqual(arms, [[1, 1], [5, 5], [2, 2]])
        self.assertEqual(
            list(gen_metadata["regression"]), [False, False, True, True, False]
        )
        self.assertTrue(gen_metadata["prob_infeasibility"][0, 0] < 0.9)
        self.assertTrue(gen_metadata["prob_infeasibility"][1, 0] < 0.9)
        self.assertTrue(gen_metadata["prob_infeasibility"][2, 1] > 0.9)
        self.assertTrue(gen_metadata["prob_infeasibility"][3, 0] > 0.9)
        self.assertTrue(gen_metadata["prob_infeasibility"][4, 0] < 0.9)
        self.assertEqual(weights, [1.0] * 3)
        self.assertEqual(gen_metadata["best_x"], arms[0])

        with self.subTest("No variance"):
            generator.fit(
                Xs=self.Xs,
                Ys=self.Ys,
                Yvars=[[0] * len(self.Ys[0]) for _ in self.Ys],
                parameter_values=self.parameter_values,
                outcome_names=self.outcome_names,
            )
            arms, weights, gen_metadata = generator.gen(
                n=3,
                parameter_values=self.parameter_values,
                outcome_constraints=(np.array([[0.0, -1.0]]), np.array([[0.0]])),
                objective_weights=np.array([1, 0]),
            )
            self.assertEqual(sorted([tuple(a) for a in arms]), [(1, 1), (2, 2), (5, 5)])

            self.assertEqual(
                list(gen_metadata["regression"]), [True, True, True, True, False]
            )
            prob_infeasibility = gen_metadata["prob_infeasibility"]
            self.assertEqual(prob_infeasibility[0, 0], 0)
            self.assertEqual(prob_infeasibility[1, 0], 1)
            self.assertEqual(prob_infeasibility[2, 1], 1)
            self.assertEqual(prob_infeasibility[3, 0], 1)
            self.assertEqual(prob_infeasibility[4, 0], 0)

            self.assertEqual(weights, [1.0] * 3)
            self.assertEqual(gen_metadata["best_x"], arms[0])

        with catch_warnings(record=True) as warning_list:
            arms, weights, gen_metadata = generator.gen(
                n=3,
                parameter_values=self.parameter_values,
                outcome_constraints=None,
                objective_weights=np.array([1, -1]),
            )
        self.assertEqual(
            "In case of multi-objective adding metric values together might"
            " not lead to a meaningful result.",
            str(warning_list[0].message),
        )

    def test_no_model_feasibility_util(self) -> None:
        probabilities = no_model_feasibility_util(
            lb=1.0, ub=2.0, Y=np.array([0.0, 1.0, 1.5, 2.0, 3.0])
        )
        self.assertTrue((probabilities == [False, True, True, True, False]).all())



# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

# pyre-strict


from unittest.mock import patch

from ax.benchmark.benchmark_problem import BenchmarkProblem
from ax.benchmark.problems.registry import (
    BENCHMARK_PROBLEM_REGISTRY,
    get_benchmark_problem,
)
from ax.benchmark.problems.runtime_funcs import int_from_params
from ax.benchmark.testing.benchmark_stubs import get_mock_lcbench_data
from ax.utils.common.testutils import TestCase
from sklearn.pipeline import Pipeline


class TestProblems(TestCase):
    def test_load_problems(self) -> None:
        # Make sure problem construction succeeds
        for name in BENCHMARK_PROBLEM_REGISTRY.keys():
            if "MNIST" in name:
                continue  # Skip these as they cause the test to take a long time

            # Avoid downloading data from the internet
            with patch(
                "ax.benchmark.problems.surrogate."
                "lcbench.early_stopping.load_lcbench_data",
                return_value=get_mock_lcbench_data(),
            ), patch.object(Pipeline, "fit"):
                problem = get_benchmark_problem(problem_key=name)
            self.assertIsInstance(problem, BenchmarkProblem, msg=name)

    def test_name(self) -> None:
        expected_names = [
            ("Bandit", "Bandit"),
            ("branin", "Branin"),
            ("hartmann3", "Hartmann_3d"),
            ("hartmann6", "Hartmann_6d"),
            ("hartmann30", "Hartmann_30d"),
            ("branin_currin_observed_noise", "BraninCurrin_observed_noise"),
            ("branin_currin30_observed_noise", "BraninCurrin_observed_noise_30d"),
            ("levy4", "Levy_4d"),
        ] + [
            (name, name)
            for name in ["Discrete Ackley", "Discrete Hartmann", "Discrete Rosenbrock"]
        ]
        for registry_key, problem_name in expected_names:
            problem = get_benchmark_problem(problem_key=registry_key)
            self.assertEqual(problem.name, problem_name)

    def test_no_duplicates(self) -> None:
        keys = [elt for elt in BENCHMARK_PROBLEM_REGISTRY.keys() if "MNIST" not in elt]

        # Avoid downloading data from the internet
        with patch(
            "ax.benchmark.problems.surrogate."
            "lcbench.early_stopping.load_lcbench_data",
            return_value=get_mock_lcbench_data(),
        ), patch.object(Pipeline, "fit"):
            names = {get_benchmark_problem(problem_key=key).name for key in keys}

        self.assertEqual(len(keys), len(names))

    def test_external_registry(self) -> None:
        registry = {
            "test_problem": BENCHMARK_PROBLEM_REGISTRY["branin"],
        }
        problem = get_benchmark_problem(problem_key="test_problem", registry=registry)
        self.assertEqual(problem.name, "Branin")
        with self.assertRaises(KeyError):
            get_benchmark_problem(problem_key="branin", registry=registry)

    def test_registry_kwargs_not_mutated(self) -> None:
        entry = BENCHMARK_PROBLEM_REGISTRY["jenatton"]
        expected_kws = {"num_trials": 50, "observe_noise_sd": False}
        self.assertEqual(entry.factory_kwargs, expected_kws)
        problem = get_benchmark_problem(problem_key="jenatton", num_trials=5)
        self.assertEqual(problem.num_trials, 5)
        self.assertEqual(
            BENCHMARK_PROBLEM_REGISTRY["jenatton"].factory_kwargs, expected_kws
        )
        problem = get_benchmark_problem(problem_key="jenatton")
        self.assertEqual(problem.num_trials, 50)

    def test_runtime_funcs(self) -> None:
        parameters = {"x0": 0.5, "x1": -3, "x2": "-4", "x3": False, "x4": None}
        result = int_from_params(params=parameters)
        expected = 1
        self.assertEqual(result, expected)
